"""Class definition of the Scipy piecewise exponential propagation model.

Uses the GRAPE optimization method.
Assumes that the signal is piecewise constant (PWC) without an LO and the
Hamiltonian is defined in the rotating frame of drive.

"""

from functools import partial

import jax
import jax.numpy as jnp
from jax import jit, vmap
from jax.lax import scan
from jax.scipy.linalg import expm, expm_frechet

from paraqeet.exceptions import ConfigurationException
from paraqeet.model.equation_of_motion import EquationOfMotion
from paraqeet.propagation.differentiable_propagation import DifferentiablePropagation
from paraqeet.propagation.scipy_expm import ScipyExpm
from paraqeet.quantity import Array

jax.config.update("jax_enable_x64", True)


class ScipyExpmGRAPE(ScipyExpm, DifferentiablePropagation):
    """Solve EOMs by piecewise exponentiation via Scipy using GRAPE.

    Compute the gradients of a closed quantum system for PWC pulses by using
    GRAPE. Here, we use forward propagation of the initial state and backward
    propagation of the target state to compute the gradients.

    The state propagations are done by the `ScipyExpm` method.

    _res: float
        Simulation resolution.
    _initial_state: Array = None
        Initial state for forward propagation.
    _target_state: Array = None
        Target state for backward propagation.
    _schirmer_derivative: bool = False
        If true, compute the gradient by Schirmer Derivative/Method of auxiliary
        matrix exponential. If false, use frechet derivative.
    """

    # TODO: Add internal time-list as an attribute that stores the times to evaluate the pulse at

    _target_state: Array | None = None
    _schirmer_derivative: bool = False

    def __init__(self, model: EquationOfMotion, resolution: float):
        super().__init__(model, resolution)

    def set_target_state(self, target_state: Array) -> None:
        """Set target state for backward propagation.

        Parameters
        ----------
        target_state: Array
            Target state.
        """
        # Verify if `model.ode_propagation` is set to `False`.
        # ode_propgation returns hamiltonian and collapse operators separately.
        if self._model is None:
            raise ConfigurationException("No equation of motion is configured.")
        eom = self._model.get_value(jnp.array([0]))
        if len(eom) == 2:
            raise ConfigurationException("Please set `model.ode_propagation` to `False` for this propagation method.")

        # For open system convert Density Matrix to Vectorized form.
        if self.is_open:
            try:
                if len(target_state.shape) == 1:  # An (n,) array
                    target_state = jnp.reshape(target_state, (-1, 1))
                # Compare the shapes of target state with the generator of time translation
                dim_generator = eom.shape[1]
                # Comparing dim -2 as 0 can be batch dimension
                if target_state.shape[-2] == jnp.sqrt(dim_generator):
                    # check if it is a square matrix. Check the last 2 dimensions are equal.
                    if target_state.shape[-1] == target_state.shape[-2]:
                        # This is a density matrix
                        target_state = ScipyExpm._convert_dm_to_vec(target_state, dim_generator)
            except Exception as e:
                raise ConfigurationException(
                    f"Obtained a state vector of shape {target_state.shape} as target state. "
                    + "For open system propagation expected a density matrix or vectorized density matrix "
                    + "as the target state.\n"
                    + f"Raised exception: `{e}`"
                )
        self._target_state = target_state

    @property
    def use_schirmer_derivative(self) -> bool:
        """Returns whether the Schirmer method is used to compute the derivative of the unitary operator."""
        return self._schirmer_derivative

    @use_schirmer_derivative.setter
    def use_schirmer_derivative(self, schirmer_derivative: bool) -> None:
        """Schirmer Derivative method to compute derivative of Unitary operator.

        Parameters
        ----------
        schirmerDerivative : bool
            If True use Schirmer derivative, if False use Frechet Derivative.
        """
        self._schirmer_derivative = schirmer_derivative

    @staticmethod
    @jit
    def _sandwich_op_values(
        bwd_propagated_state: Array,
        op: Array,
        fwd_propagated_state: Array,
    ) -> Array:
        r"""Compute \\langle \\lambda(t) | O | \\psi(t) \\rangle.

        Parameters
        ----------
        bwd_propagated_state: Array
            Backwards propagated states
        op: Array
            Array of operator for each time point.
        fwd_propagated_state: Array
            Forwards propagated states

        Returns
        -------
        Array
            Matrix element of the operator for each time point.
        """
        return jnp.matmul(bwd_propagated_state, jnp.matmul(op, fwd_propagated_state))

    @partial(jit, static_argnums=(0,))
    def _forward_and_backward_propagation(
        self,
        us,
        psis_t,
        lamdas_t,
        steps_arr,
    ):
        """Forward propagate initial state and backward propagate target state.

        JIT compiled and uses `jax.lax.scan` to avoid compilation overhead.

        Parameters
        ----------
        psis_t : Array
            Forward propagated state
        lamdas_t : Array
            Backward propagated state
        """

        def forward_propagation(psis_t, index):
            psis_t = us[index] @ psis_t
            return psis_t, psis_t

        def backward_propagation(lamdas_t, index):
            lamdas_t = lamdas_t @ us[-index - 1]
            return lamdas_t, lamdas_t

        psis_t, psis_list = scan(forward_propagation, psis_t, steps_arr)
        lamdas_t, lamdas_list = scan(backward_propagation, lamdas_t, steps_arr)

        return psis_list, lamdas_list

    @partial(jit, static_argnums=(0,))
    def _forward_and_backward_propagation_open(
        self,
        us,
        us_rev,
        psis_t,
        lamdas_t,
        steps_arr,
    ):
        """Forward propagate initial state and backward propagate target state.

        JIT compiled and uses `jax.lax.scan` to avoid compilation overhead.

        Parameters
        ----------
        psis_t: Array
            Forward propagated state
        lamdas_t: Array
            Backward propagated state
        """

        def forward_propagation(psis_t, index):
            psis_t = us[index] @ psis_t
            return psis_t, psis_t

        def backward_propagation(lamdas_t, index):
            lamdas_t = us_rev[index] @ lamdas_t
            return lamdas_t, lamdas_t

        psis_t, psis_list = scan(forward_propagation, psis_t, steps_arr)
        lamdas_t, lamdas_list = scan(backward_propagation, lamdas_t, steps_arr)

        return psis_list, lamdas_list

    @staticmethod
    @partial(jit, static_argnums=(0,))
    def _exponentiate_frechet(dim, ham, dh_dp):
        r"""Exponentiate and also calculate the frechet derivative.

        Parameters
        ----------
        ham: Array
            -iHdt
        dh_dp: Array
            -i\\frac{\\partial H}{\\partial u} dt
        """
        return expm_frechet(ham, dh_dp)

    @staticmethod
    @partial(jit, static_argnums=(0,))
    def _exponentiate_schirmer(dim, ham, dh_dp):
        r"""Exponentiate an auxiliary matrix to compute U and dU.

        Parameters
        ----------
        ham : Array
            -iHdt
        dh_dp : Array
            -i\\frac{\\partial H}{\\partial u} dt
        """
        zeros = jnp.zeros_like(ham)
        h_extended = jnp.block([[ham, dh_dp], [zeros, ham]])
        u_extended = expm(h_extended)
        return u_extended[:dim, :dim], u_extended[:dim, dim:]

    @staticmethod
    @jit
    def _exponentiate(ham):
        r"""Exponentiate EOM using Expm.

        Parameters
        ----------
            ham : Array
            -iHdt
        """
        return expm(ham)

    @partial(jit, static_argnums=(0,))
    def _propagate_in_time(
        self,
        us,
        psis_t,
        steps_arr,
    ):
        """Propagate Full time.

        JIT compiled and uses `jax.lax.scan` to avoid compilation overhead.

        Parameters
        ----------
        psis_t : Array
            Forward propagated state
        lamdas_t : Array
            Backward propagated state
        """

        def forward_propagation(psis_t, index):
            psis_t = us[index] @ psis_t
            return psis_t, psis_t

        psis_t, psis_list = scan(forward_propagation, psis_t, steps_arr)
        return psis_list

    def propagate(self, time: Array) -> Array:
        """Loop over all desired times in time at set resolution."""
        if len(time) < 2:
            raise ValueError("ScipyExpmGRAPE.propagate needs at least two time points.")

        if self._initial_state is None:
            raise ConfigurationException("Initial state is not set")

        init_state = jnp.array(self._initial_state, dtype=jnp.complex128)
        dt = time[1] - time[0]

        time_grid = time[:-1] + dt / 2

        if self._model is None:
            raise ConfigurationException("No model is configured to provide an equation of motion.")

        eom_func = self._model.get_value
        eom = eom_func(time_grid) * dt

        us = vmap(ScipyExpmGRAPE._exponentiate, in_axes=(0,))(eom)

        psis = self._propagate_in_time(us, init_state, jnp.arange(0, len(time_grid), 1))
        psis = jnp.concat([jnp.expand_dims(init_state, axis=0), psis], axis=0)

        # if open system convert back the vectorized density matrices to matrix shape
        dim = eom.shape[-2]
        if self.is_open:
            psis = jnp.array(psis)
            psis = vmap(ScipyExpm._convert_vec_to_dm, in_axes=(0, None))(psis, int(jnp.sqrt(dim)))
        return jnp.array(psis)

    def _gradient_closed_system(self, time: Array) -> tuple[Array, Array]:
        init_state = jnp.array(self._initial_state, dtype=jnp.complex128)
        target_state = jnp.array(self._target_state, dtype=jnp.complex128)
        target_state = target_state.conj().T

        if self._model is None:
            raise ConfigurationException("No model is configured to provide an equation of motion.")

        grad_func = self._model.get_value_and_gradient

        dt = time[1] - time[0]

        time_grid = time[:-1] + dt / 2

        hams, dh_dps = grad_func(time_grid)
        hams = hams * dt
        dh_dps = jnp.array(dh_dps) * dt

        u_grads_list = []
        n_params = dh_dps.shape[1]

        dim = hams.shape[-2]

        if self._schirmer_derivative:
            exponentiating_function = ScipyExpmGRAPE._exponentiate_schirmer
        else:
            exponentiating_function = ScipyExpmGRAPE._exponentiate_frechet

        for i in range(n_params):
            us, d_us = vmap(exponentiating_function, in_axes=(None, 0, 0))(dim, hams, dh_dps[:, i, ...])
            u_grads_list.append(d_us)

        u_grads = jnp.stack(u_grads_list, axis=1)

        psis, lamdas = self._forward_and_backward_propagation(
            us, init_state, target_state, jnp.arange(0, len(time_grid), 1)
        )

        psis = jnp.concat([jnp.expand_dims(init_state, axis=0), psis], axis=0)
        lamdas = jnp.concat([jnp.expand_dims(target_state, axis=0), lamdas], axis=0)

        lamdas = jnp.flip(lamdas, axis=0)

        grads = []
        for i in range(n_params):
            grad = vmap(
                ScipyExpmGRAPE._sandwich_op_values, in_axes=(0, 0, 0)
            )(
                lamdas[1:],
                u_grads[:, i, ...],  # type: ignore
                psis[:-1],
            )
            grad = jnp.squeeze(grad)
            grads.append(grad)

        return psis, jnp.array(grads)

    def _gradient_open_systems(self, time: Array) -> tuple[Array, Array]:
        raise NotImplementedError(
            "Currently ScipyExpmGRAPE is not supported for open system optimization."
            + " Use Vern7GRAPE as an alternative (with `model.ode_propagation = True`)."
        )

    def get_value_and_gradient(self, times: Array) -> tuple[Array, Array]:
        """Compute gradients using GRAPE.

        Compute the forward propagation of the initial state and
        the backward propagation of the target state.

        Psis represent the forward propagation and lamdas represent
        the backward propagation states.

        This propagation method assumes a PWC pulse as input.
        """
        if len(times) < 2:
            raise ValueError("ScipyExpmGRAPE.get_value_and_gradient needs at least two time points.")

        if self._initial_state is None:
            raise ConfigurationException("Initial state is not set")

        if self._target_state is None:
            raise ConfigurationException("Target state is not set")

        if self.is_open:
            psis, grads = self._gradient_open_systems(times)
        else:
            psis, grads = self._gradient_closed_system(times)

        return psis, grads



"""Class definition of the Bayesian Optimizer model."""

import jax.numpy as jnp
from bayes_opt import BayesianOptimization

from paraqeet.exceptions import ConfigurationException
from paraqeet.measurement.measurement import NormalizableMeasurement
from paraqeet.optimization_map import OptimizationMap
from paraqeet.optimizers.optimizer import OptimizationResult, Optimizer
from paraqeet.quantity import Array


class BayesianOptimizer(Optimizer):
    """Minimizes the outcome of a measuremnt using Bayesian optimization.

    This is useful if the evaluation of the measurement is costly.
    This class is mostly a wrapper around the implementing package.

    See Also
    --------
    https://bayesian-optimization.github.io/BayesianOptimization/index.html

    Parameters
    ----------
    measure: Measurement
        The measure to be optimized.
    optimization_map : OptimizationMap
        All optimizable parameters via the optimization map.
    initial_samples : int=10
        Number of iterations before the explorations starts the exploration
        for the maximum.
    iterations: int =100
        Number of iterations where the method attempts to find the maximum
        value.

    """

    _measure: NormalizableMeasurement
    _initial_samples: int
    _iterations: int

    def __init__(
        self,
        measure: NormalizableMeasurement,
        optimization_map: OptimizationMap,
        initial_samples=10,
        iterations=100,
    ):
        super().__init__(measure, optimization_map)
        self._initial_samples = initial_samples
        self._iterations = iterations

    @property
    def initial_samples(self) -> int:
        """Get the initial samples fed to the system."""
        return self._initial_samples

    @initial_samples.setter
    def initial_samples(self, initial_samples: int) -> None:
        """Set the initial samples for the system."""
        self._initial_samples = initial_samples

    @property
    def iterations(self) -> int:
        """Get the iterations of the system."""
        return self._iterations

    @iterations.setter
    def iterations(self, iterations: int) -> None:
        """Set the iterations of the system."""
        self._iterations = iterations

    def optimize(self, times: Array | float) -> OptimizationResult:
        """Optimize the system via the Bayesian optimizer.

        Performs the actual optimization.

        *Note - If input `times` is a float, then the start time of propagation is implicitly assumed to be zero.
        For an array of times, the first time point is the start time.*

        Returns
        -------
        OptimizationResult
            Result of optimization via the OptimizationResult object.
            (status, value, iterations and the raw result)

        """
        self._times = jnp.array([0.0, times]) if isinstance(times, float) else times

        if self._logger:
            self._logger.start()

        self._optimization_map.register_params_with_optimizables()
        params = self._optimization_map.get_all_parameters()

        # The optimizer needs a dict of named bounds. We use the parameters'
        # indices in the list as names because the parameters' names might
        # not be unique. The bounds are in the reduced representation
        # because this will be the working range for the optimizer.
        optimizer = BayesianOptimization(
            f=self._set_parameters_and_measure,
            pbounds={str(i): (-1, 1) for i in range(len(params))},
            verbose=2,
            random_state=1,
        )
        optimizer.maximize(init_points=self._initial_samples, n_iter=self._iterations)

        # The last measurement is not necessarily the best.
        # We therefore set the optimized parameters to the best value.
        if not optimizer.max:
            raise ConfigurationException("BayesianOptimization has no max field.")

        best_values = optimizer.max["params"]
        for i, param in enumerate(params):
            param.set_reduced_value(best_values[str(i)])

        # Use the actual names and the non-reduced values
        # for the return value
        result = {params[i].get_name(): params[i].get_value() for i in range(len(best_values))}
        result["fun"] = 1 - optimizer.max["target"]
        if self._logger:
            self._logger.stop(str(result))

        return OptimizationResult(
            status=OptimizationResult.STATUS_FINISHED,
            value=float(result["fun"]),
            iterations=self._iterations + self._initial_samples,
            raw_result=optimizer.max,
        )

    def _set_parameters_and_measure(self, **kwargs) -> float:
        """Update the parameter values and returns the measurement result.

        Internal callback.

        Parameters
        ----------
        **kwargs
            A dict mapping parameter names to their values.

        Returns
        -------
        Array
            Returns the fidelity after setting the parameters.

        """
        log = []
        params = self._optimization_map.get_all_parameters()
        for i, param in enumerate(params):
            param.set_reduced_value(kwargs[str(i)])
            log.append(params[i])

        fidelity = self._measure.calculate_normalized_scalar(self._times)

        if self._logger:
            self._logger.log(log, fidelity)
        return fidelity



"""The class definition of state transfer fidelity model."""

import warnings
from collections.abc import Callable

import jax
import jax.numpy as jnp
from jax import grad, jit

from paraqeet.differentiable import Differentiable
from paraqeet.measurement.measurement import NormalizableMeasurement
from paraqeet.propagation.differentiable_propagation import DifferentiablePropagation
from paraqeet.quantity import Array

jax.config.update("jax_enable_x64", True)


class StateTransferFidelity(NormalizableMeasurement, Differentiable):
    """Fidelity measure that compares overlap of the initial and final state.

    Parameters
    ----------
    propagation : DifferentiablePropagation
        Abstract base class for any implementation that can solve
        the equation of motion.
    initial_state : Array
        Initial state.
    target_state : Array
        Target state.
    times : Array
        One-dimensional vector of timestamps.

    """

    _initial_state: Array
    _target_state: Array
    _propagation: DifferentiablePropagation

    def __init__(
        self,
        propagation: DifferentiablePropagation,
        initial_state: Array,
        target_state: Array,
    ):
        self._propagation = propagation
        self._initial_state = initial_state
        self._target_state = target_state
        if target_state.shape != initial_state.shape:
            warnings.warn(
                UserWarning(
                    f"Different shapes for target_state({target_state.shape})"
                    f"and initial_state({initial_state.shape}) detected."
                    " Use restrict_subsystems to project states to "
                    "the same shape before measuring."
                )
            )
        self._propagation.set_initial_state(self._initial_state)
        if self._propagation.is_open:
            self._overlap = self._overlap_dm
        else:
            self._overlap = self._overlap_vec

    @staticmethod
    def _fid(overlap: Array) -> float:
        return float(jnp.abs(jnp.average(overlap)) ** 2)

    @staticmethod
    def _overlap_vec(target_state, final_state):
        return jnp.vdot(target_state, final_state)

    @staticmethod
    def _overlap_dm(target_state, final_state):
        return jnp.linalg.trace(jnp.matmul(target_state, final_state))

    def measure(self, times: Array) -> Array | float:
        """Return measurement in the range [0, 1]."""
        return self.calculate_normalized_scalar(times=times)

    def calculate_normalized_scalar(self, times: Array | float) -> float:
        """Measure overlap between initial and target state. To be used with an optimizer.

        Parameters
        ----------
        times : Array
            One-dimensional vector of timestamps.

        Returns
        -------
        float
            Overlap between initial and target state in a bare float.

        """
        # TODO: `propagate` needs at least two time points initial and final.
        # TODO: Does Measurement implement default conversion from times: float -> Array?
        # TODO: Fix typing
        states = self._propagation.propagate(time=times)
        final_state = states[-1]
        f = self._overlap(self._target_state, final_state)
        return StateTransferFidelity._fid(f)

    def get_value_and_gradient(self, times: Array) -> tuple[Array, Array] | tuple[float, Array]:
        """Compute function value and corresponding gradient.

        Returns
        -------
        Tuple[Array, Array]
            Tuple of function value and gradient of shape (n_parameters,).

        """
        states, dg_dp_list = self._propagation.get_value_and_gradient(times=times)
        final_state = states[-1]
        df_dp_list = []
        f = self._overlap(self._target_state, final_state)
        for dg_dp in dg_dp_list[-1]:
            g = self._overlap(self._target_state, dg_dp)
            df_dp_list.append(jnp.real(f.conj() * g + f * g.conj()))  # chain rule for abs^2
        return StateTransferFidelity._fid(f), jnp.array(df_dp_list)  # shape scalar, (n_parameters,)


class StateTransferFidelityAD(StateTransferFidelity):
    """Fidelity measure that compares overlap of the initial and final state.

    Parameters
    ----------
    propagation : Propagation
        Abstract base class for any implementation that can solve
        the equation of motion.
    initial_state : Array
        Initial state.
    target_state : Array
        Target state.
    times : Array
        One-dimensional vector of timestamps.

    """

    _gradient_function: Callable | None

    def __init__(
        self,
        propagation: DifferentiablePropagation,
        initial_state: Array,
        target_state: Array,
    ):
        super().__init__(propagation, initial_state, target_state)
        self._gradient_function = None

    def get_value_and_gradient(self, times: Array) -> tuple[Array, Array] | tuple[float, Array]:
        """Measure with gradient.

        Overwrite inherited `measure_with_gradient` to calculate
        gradients using AD.

        Returns
        -------
        Tuple[float, Array]
            Tuple of function value and gradient of shape (n_parameters,).

        """
        if self._gradient_function is None:
            self._gradient_function = jit(grad(self._fid, argnums=0))

        states, dg_dp_list = self._propagation.get_value_and_gradient(times=times)
        final_state = states[-1]
        df_dp_list = []
        f = self._overlap(self._target_state, final_state)
        for dg_dp in dg_dp_list[-1]:
            g = self._overlap(self._target_state, dg_dp)
            dfdp = self._gradient_function(f) * g
            df_dp_list.append(jnp.real(dfdp))
        return StateTransferFidelity._fid(f), jnp.array(df_dp_list)  # shape scalar, (n_parameters,)


class StateTransferFidelityGRAPE(StateTransferFidelity):
    """Fidelity measure that compares overlap of the initial and final state.

    For GRAPE the optimizable parameters are vector quantities.

    Parameters
    ----------
    propagation : DifferentiablePropagation
        Abstract base class for any implementation that can solve
        the equation of motion.
    initial_state : Array
        Initial state.
    target_state : Array
        Target state.
    times : Array
        One-dimensional vector of timestamps.

    """

    _propagation: DifferentiablePropagation

    def get_value_and_gradient(self, times: Array) -> tuple[Array, Array] | tuple[float, Array]:
        """Compute function value and corresponding gradient.

        Returns
        -------
        Tuple[Array, Array]
            Tuple of function value and gradient of shape (n_parameters,).

        """
        states, grads = self._propagation.get_value_and_gradient(times=times)
        final_state = states[-1]
        f = self._overlap(self._target_state, final_state)
        if self._propagation.is_open:
            grads = jnp.real(jnp.linalg.trace(grads)).flatten()
        else:
            grads = 0.5 * jnp.real(f.conj() * grads + grads.conj() * f).flatten()
        return StateTransferFidelity._fid(f), grads  # shape scalar, (n_parameters,)



"""Class definition for the Runge-Kutta Scipy propagation model."""

import numpy as np  # Using regular numpy for scipy interface
from scipy.integrate import RK45  # TODO: Replace with jax? Is there one?

from paraqeet.exceptions import ConfigurationException
from paraqeet.model.equation_of_motion import EquationOfMotion
from paraqeet.propagation.propagation import Propagation
from paraqeet.quantity import Array, Quantity


class RungeKutta(Propagation):
    """Propagation via the Runge-Kutta Scipy implementation.

    Uses scipy's Runge-Kutta implementation for propagating
    a state vector or density matrix.

    Parameters
    ----------
    model : Model
        Represents the equation of motion for a given Hamiltonian.
    initial_time_step : float | None, optional
        The initial time step for the adaptive time steps in RK45.

    """

    _initial_time_step: float | None

    def __init__(self, model: EquationOfMotion, initial_time_step: float | None = None):
        # TODO: setting a default value for resolution
        initial_time_step = 0.1e-9 if initial_time_step is None else initial_time_step
        super().__init__(model, resolution=1 / initial_time_step)
        self._initial_state: Array
        self._initial_time_step = initial_time_step

    def get_parameters(self) -> list[Quantity]:
        """Get a list of parameters of the system.

        Returns
        -------
        list[Quantity]
            List of optimizable parameters of the system.

        """
        return []

    def set_initial_state(self, state: Array):
        """Set the initial state for the propagation.

        Subclasses can access the state in the _initial_state field.

        Parameters
        ----------
        state: Array
            Parameter value to be set as the initial state for the propagation.

        """
        self._initial_state = np.reshape(state, (-1,))

    def propagate(self, time: Array) -> Array:
        """Return the solution of the equations of motion.

        Parameters
        ----------
        time: Array
            Any one-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the solution of the equations of motion.

        Raises
        ------
        ConfigurationException
            If the initial state is not set.
        ValueError
            If the propagation needs at least two time steps.

        """
        if self._initial_state is None:
            raise ConfigurationException("Initial state is not set")

        if len(time) < 2:
            raise ValueError("RungeKutta.propagate needs at least two time steps")

        def callback(time, state):
            column_state = np.reshape(state, (-1, 1))
            return np.reshape(
                self._model.get_right_hand_side(np.array([time]), column_state),
                (-1,),
            )

        # Since RK45 uses adaptive time steps and does not guarantee
        # to return a state for each time stamp, this
        # function has to iterate over the time steps itself.
        states = [self._initial_state]
        for ti in range(1, len(time)):
            dt = self._initial_time_step
            if dt is None or dt > time[ti] - time[ti - 1]:
                dt = float(time[ti] - time[ti - 1]) / 5

            # This is the scipy implementation of RK45, which is compatible with (non-jax) numpy
            integrator = RK45(
                fun=callback,
                t0=time[ti - 1],
                y0=np.reshape(states[-1], (-1,)),
                t_bound=time[ti],
                first_step=dt,
                vectorized=False,
            )

            while integrator.status == "running":
                integrator.step()
            states.append(integrator.y)
        return np.array(states)



# Helper functions that are common tests for all propagation implementations
import numpy as np
import pytest

from paraqeet.exceptions import ConfigurationException
from paraqeet.propagation.propagation import Propagation


def needs_initial_state(propagation: Propagation, dimension: int):
    random_time_vector = np.linspace(0.0, np.random.randint(1, 10) * np.random.rand(), np.random.randint(2, 10))
    with pytest.raises(ConfigurationException):
        propagation.propagate(random_time_vector)

    state = np.random.random(dimension) + 1j * np.random.random(dimension)
    state = state / np.sqrt(np.vdot(state, state))
    propagation.set_initial_state(state)
    propagation.propagate(random_time_vector)



"""Test the Scipy piecewise exponentiation GOAT solver."""

import numpy as np
import pytest

from paraqeet.propagation.scipy_expm_goat import ScipyExpmGOAT
from tests.model.dummy_model import DummyEquationsOfMotion
from tests.model.empty_hamiltonian import EmptyHamiltonian
from tests.propagation.test_common_propagation import needs_initial_state


@pytest.fixture
def expm():
    """Return a Scipy piecewise exponentiation solver generating function."""

    def _method(dimension, resolution):
        return ScipyExpmGOAT(DummyEquationsOfMotion(EmptyHamiltonian(dimension)), resolution=resolution)

    return _method


def test_parameters(expm):
    """Test parameters from the equations of motion."""
    propagation = expm(dimension=np.random.randint(10), resolution=3)
    assert propagation.get_parameters() == []


def test_resolution(expm):
    """Test the resolution of the solver."""
    for _ in range(10):
        propagation = expm(dimension=np.random.randint(2, 100), resolution=3)
        resolution = np.random.randint(1, 1000)
        propagation.resolution = resolution
        assert propagation.resolution == resolution


def test_state_dimension_vector(random_state, expm, ts):
    """Test dimension and norm of state vectors after propagation.

    The dimension and norm of state vectors should be the same
    after propagation.

    """
    for _ in range(10):
        dim = np.random.randint(2, 10)
        state = random_state(dim)
        propagation = expm(dim, resolution=3)
        propagation.set_initial_state(state)
        propagated_states = propagation.propagate(ts)
        assert propagated_states.shape[0] == len(ts)
        assert propagated_states.shape[1:] == state.shape + (1,)


@pytest.mark.parametrize("is_open", [True, False])
def test_state_dimension_matrix(random_matrix, expm, ts, is_open):
    """Test the state matrix after propagation."""
    for _ in range(10):
        dim = np.random.randint(2, 10)
        state = random_matrix(dim, dim)
        if is_open:
            propagation = expm(dim**2, resolution=3)
            propagation.is_open = True
        else:
            propagation = expm(dim, resolution=3)
        propagation.set_initial_state(state)
        propagated_states = propagation.propagate(ts)
        assert propagated_states.shape[0] == len(ts)
        assert propagated_states.shape[1:] == state.shape


def test_needs_initial_state(random_state, expm):
    for dim in range(2, 10):
        propagation = expm(dim, resolution=3)
        needs_initial_state(propagation, dim)



"""Class definition of the fixed time-step 7th-order Verner ODE solver. Adapted from Julia DiffEq Vern7."""

from functools import partial

import jax
import jax.numpy as jnp
from jax import jit
from jax.lax import dynamic_slice_in_dim, scan

from paraqeet.exceptions import ConfigurationException
from paraqeet.model.equation_of_motion import EquationOfMotion
from paraqeet.propagation.propagation import Propagation
from paraqeet.quantity import Array

jax.config.update("jax_enable_x64", True)


class Vern7(Propagation):
    """
    Propagate state by solving the Schrödinger equation / Lindblad master equation by using ODE solver.

    Implements Vern7 ODE Solver algorithm non adaptive (fixed time-step) version.
    """

    _resolution: float
    _initial_state: Array | None = None

    def __init__(self, model: EquationOfMotion, resolution: float):
        """
        Parameters
        ----------
        model: Model
            Model
        res: float
            Resolution at which to sample the EOM
        """
        super().__init__(model, resolution)
        self.resolution = resolution

        if self.is_open:
            self.step_function = self._lindblad_step
        else:
            self.step_function = self._schrodinger_step

    def set_initial_state(self, state):
        """Set initial state."""
        # For open system check if initial state is a density matrixs.
        if self.is_open:
            if state.shape[-1] != state.shape[-2]:
                raise ConfigurationException(
                    f"Obtained a state vector of shape {state.shape} as initial state. "
                    + "For open system propagation expected a density matrix as the initial state."
                )
        self._initial_state = jnp.array(state, dtype=jnp.complex128)

    @staticmethod
    def _commutator(A: Array, B: Array):
        return jnp.matmul(A, B) - jnp.matmul(B, A)

    @staticmethod
    def _anti_commutator(A: Array, B: Array):
        return jnp.matmul(A, B) + jnp.matmul(B, A)

    @staticmethod
    def _dagger(op: Array):
        return op.conj().T

    @staticmethod
    def _interpolate_time(times, dt):
        times_interp = jnp.concatenate(
            [
                times,
                times + (1 / 200) * dt,
                times + (49 / 450) * dt,
                times + (49 / 300) * dt,
                times + (911 / 2000) * dt,
                times + (3480084980 / 5709648941) * dt,
                times + (221 / 250) * dt,
                times + (37 / 40) * dt,
                times + dt,
            ],
            axis=0,
        )
        return jnp.sort(times_interp)

    def _lindblad_step(self, state: Array, h: Array, cols: list[Array]):
        del_rho = Vern7._commutator(h, state)
        for col in cols:
            del_rho += jnp.matmul(jnp.matmul(col, state), Vern7._dagger(col))
            del_rho -= 0.5 * Vern7._anti_commutator(jnp.matmul(Vern7._dagger(col), col), state)
        return del_rho

    def _schrodinger_step(self, state: Array, h: Array, cols: list[Array]):
        return jnp.matmul(h, state)

    @partial(jit, static_argnums=(0,))
    def _vern7_one_step(self, state, h, col):
        k1 = self.step_function(state, h[0], col)
        k2 = self.step_function(state + (1 / 200) * k1, h[1], col)
        k3 = self.step_function(state + (-4361 / 4050) * k1 + (2401 / 2025) * k2, h[2], col)
        k4 = self.step_function(
            state + (49 / 1200) * k1 + (49 / 400) * k3,
            h[3],
            col,
        )
        k5 = self.step_function(
            state + (2454451729 / 3841600000) * k1 + (-9433712007 / 3841600000) * k3 + (4364554539 / 1920800000) * k4,
            h[4],
            col,
        )
        k6 = self.step_function(
            state
            + (-6187101755456742839167388910402379177523537620 / 2324599620333464857202963610201679332423082271) * k1
            + (27569888999279458303270493567994248533230000 / 2551701010245296220859455115479340650299761) * k3
            + (-37368161901278864592027018689858091583238040000 / 4473131870960004275166624817435284159975481033) * k4
            + (1392547243220807196190880383038194667840000000 / 1697219131380493083996999253929006193143549863) * k5,
            h[5],
            col,
        )
        k7 = self.step_function(
            state
            + (11272026205260557297236918526339 / 1857697188743815510261537500000) * k1
            + (-48265918242888069 / 1953194276993750) * k3
            + (26726983360888651136155661781228 / 1308381343805114800955157615625) * k4
            + (-2090453318815827627666994432 / 1096684189897834170412307919) * k5
            + (1148577938985388929671582486744843844943428041509 / 1141532118233823914568777901158338927629837500000)
            * k6,
            h[6],
            col,
        )
        k10 = self.step_function(
            state
            + (-511858190895337044664743508805671 / 11367030248263048398341724647960) * k1
            + (2822037469238841750 / 15064746656776439) * k3
            + (-23523744880286194122061074624512868000 / 152723005449262599342117017051789699) * k4
            + (10685036369693854448650967542704000000 / 575558095977344459903303055137999707) * k5
            + (
                -6259648732772142303029374363607629515525848829303541906422993
                / 876479353814142962817551241844706205620792843316435566420120
            )
            * k6
            + (17380896627486168667542032602031250 / 13279937889697320236613879977356033) * k7,
            h[8],
            col,
        )
        state_new = (
            state
            + (117807213929927 / 2640907728177740) * k1
            + (4758744518816629500000 / 17812069906509312711137) * k4
            + (1730775233574080000000000 / 7863520414322158392809673) * k5
            + (
                2682653613028767167314032381891560552585218935572349997
                / 12258338284789875762081637252125169126464880985167722660
            )
            * k6
            + (40977117022675781250 / 178949401077111131341) * k7
            + (2152106665253777 / 106040260335225546) * k10
        )
        return state_new

    @partial(jit, static_argnums=(0,))
    def _propagate_in_time(self, state_t, eom, col, steps_arr):
        """
        Propagate from `time[ti] to time[ti+1]`.
        JIT compiled and uses `jax.lax.scan` to avoid compilation overhead.
        """

        def propagate_body(state_t, index):
            state_t = self._vern7_one_step(
                state_t,
                dynamic_slice_in_dim(eom, start_index=9 * index, slice_size=9, axis=0),
                col,
            )
            return state_t, state_t

        state_t, _ = scan(propagate_body, state_t, steps_arr)
        return state_t

    def propagate(self, time: Array) -> Array:
        """Return the solution of the equation of motion for open/closed system using vern7 ODE solver.

        Loop over all desired times in time at set resolution.

        Parameters
        ----------
        time: Array
            Any one-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the solution of the equations of motion.

        Raises
        ------
        ConfigurationException
            If the initial state is not set.

        """
        if len(time) < 2:
            raise ValueError("Vern7.propagate needs at least two time points.")

        if self._initial_state is None:
            raise ConfigurationException("Initial state is not set")

        if len(time) < 2:
            raise ValueError("Propagation needs at least two time steps")

        init_state = jnp.array(self._initial_state, dtype=jnp.complex128)
        if self._model is None:
            raise ConfigurationException("No equation of motion is configured.")
        eom_func = self._model.get_value

        # Verify if `model.ode_propagation` is set to `True`.
        # ode_propgation returns hamiltonian and collapse operators separately.
        eom_parts = eom_func(jnp.array([0]))
        if len(eom_parts) != 2:
            raise ConfigurationException("Please set `model.ode_propagation` to `True` for this propagation method.")

        states = [init_state]

        for ti in range(1, len(time)):
            state_t = states[ti - 1]
            times, dt = self._construct_times(time, ti)
            times_interp = Vern7._interpolate_time(times, dt)
            eom, cols = eom_func(times_interp + dt / 2)
            state_t = self._propagate_in_time(
                state_t,
                eom * dt,
                jnp.array(cols) * jnp.sqrt(dt),
                jnp.arange(0, len(times), 1),
            )
            states.append(state_t)

        return jnp.array(states)



"""Testing the cosine signal generator functions."""

import numpy as np
import pytest

from paraqeet.signal.envelopes import ConstantEnvelope, ZeroEnvelope
from paraqeet.signal.iq_mixer import IQMixer

LEN_SIG = 1001


@pytest.fixture
def time_samples():
    """Generate time samples from the given signal length.

    The given signal length `LEN_SIG` is a module level global variable.

    """
    return np.linspace(0, 10e-9, LEN_SIG)


@pytest.fixture
def gen():
    """Return cosine signal generator object."""
    tone = ConstantEnvelope()
    tone.set_optimizable_parameters(tone.get_parameters())
    return IQMixer(envelopes=[tone])


@pytest.fixture
def zero_gen():
    """Return a zero tone cosine generator object."""
    tone = ZeroEnvelope()
    return IQMixer(envelopes=[tone])


@pytest.fixture
def gen_multiple_tones():
    """Return a multiple tone cosine generator."""
    tone1 = ConstantEnvelope()
    params1 = tone1.get_parameters()

    tone2 = ConstantEnvelope()
    params2 = tone2.get_parameters()
    return IQMixer(envelopes=[tone1, tone2]), np.concatenate((params1, params2))


def test_gen(gen, time_samples) -> None:
    """Computes a sample signal and checks vectorized generation."""
    sig = gen.get_value(time_samples)
    assert len(sig) == LEN_SIG


def test_zero_tone(zero_gen, time_samples) -> None:
    """Test generation of zeroTone signal."""
    sig = zero_gen.get_value(time_samples)
    assert len(sig) == LEN_SIG
    assert np.all(sig == 0)


def test_get_parameters(gen_multiple_tones):
    """Test the get parameters function.

    First four parameters are the tone parameters and last two in 'params' are
    added by the generator, i.e., phase and LO frequency.

    """
    gen, all_params = gen_multiple_tones
    params = gen.get_parameters()[:-2]
    assert np.all(params == all_params)


def test_gradient_one_time(gen):
    """Test the generate signal gradient one time function."""
    print(gen.get_parameters())
    gen.set_optimizable_parameters(gen.get_parameters())
    grads = gen.get_gradient_at_timestep(np.array([0]))
    assert grads.shape == (len(gen.get_parameters()),)


def test_gradient_shape(gen, time_samples):
    """Test the generate signal gradient function."""
    gen.set_optimizable_parameters(gen.get_parameters())
    _, grads = gen.get_value_and_gradient(time_samples)
    assert grads.shape == (time_samples.shape[0], len(gen.get_parameters()))



"""Class definition of a random measurement model for testing."""

import numpy as np

from paraqeet.measurement.measurement import NormalizableMeasurement
from paraqeet.propagation.propagation import Propagation
from paraqeet.quantity import Array


class RandomMeasurement(NormalizableMeasurement):
    """Mock class that returns a random measurement value between 0 and 1.

    Parameters
    ----------
    propagation : Propagation
        Abstract base class for any implementation
        that can solve the equation of motion.
    times : Array
        One-dimensional vector of timestamps.
    """

    _propagation: Propagation

    def __init__(self, propagation: Propagation):
        self._propagation = propagation

    # TODO: Check the implementation method measure
    def measure(self, times: Array) -> Array | float:
        return self.calculate_normalized_scalar(times=times)

    def calculate_normalized_scalar(self, times: Array | float) -> float:
        """Return the result of measurement.

        Returns
        -------
        Array
            The result of the measurement.

        """
        return float(np.random.random())



"""optimize a dCRAB pulse by a Scipy gradient based optimizer."""

import warnings

import jax.numpy as jnp
import numpy as np
from scipy.optimize import minimize

from paraqeet.differentiable import Differentiable
from paraqeet.exceptions import ConfigurationException, IncompatibleOptimizationMap
from paraqeet.measurement.measurement import NormalizableMeasurement
from paraqeet.optimization_map import OptimizationMap
from paraqeet.optimizers.optimizer import OptimizationResult, Optimizer
from paraqeet.optimizers.scipy_optimizer_gradient import ScipyOptimizerGradient
from paraqeet.quantity import Array, Quantity
from paraqeet.signal.envelopes import DCRABEnvelope
from paraqeet.signal.waveform import DRAGMixer

warnings.simplefilter("once")


# Replace the default formatwarning
def custom_formatwarning(msg, *args, **kwargs):
    """Prettier warning statements."""
    return str(msg) + "\n"


warnings.formatwarning = custom_formatwarning  # type: ignore


class DCRABOptimizerGradient(ScipyOptimizerGradient):
    """A dCRAB optimization method.

    Implements dCRAB optimization involving super-iterations that adds additional
    optimization components to the dCRAB envelope and freezes the older parameters.

    *Note - This works with a `DCRABEnvelope` or a `list[DCRABEnvelope]` as envelopes.*

    _num_iteration: int
        Current iteration number.
    _super_iteration_every: int
        Number of iterations after which one super-iteration is performed. Defaults to 30.
    _max_super_iteration_num: int
        Maximum number of super-iterations to perform. Defaults to 10.
    _num_print_every: int
        Print every this many iterations the current optimization value. Defaults to 5.
    _old_parameters_dict : dict[int, list[Quantity]]
        Store the parameters of the previous super-iteration in a dictionary labelled by the number of parameters.

    Parameters
    ----------
    measure: NormalizableMeasurement
        Measurement class that measures the observable to be maximised.
    optimization_map: OptimizationMap
        An optimization map containing all parameters that can be optimized.
    """

    _num_iteration: int = 0
    _super_iteration_every: int
    _max_super_iteration_num: int
    _num_print_every: int
    _old_parameters_dict: dict[int, list[Quantity]]
    _old_opt_idxs_dict: dict[int, list[int]]

    best_params: list[float]
    best_fid: float = 99999

    _previous_fid: float = 99999
    _super_iteration_since: int = 0
    _super_iteration_num: int = 0
    _fallback_optimization: Optimizer | None
    _super_iteration_tol: float
    _seed: int | None = None

    def __init__(
        self,
        measure: NormalizableMeasurement,
        optimization_map: OptimizationMap,
        super_iteration_every: int = 30,
        max_super_iteration_num: int = 10,
        print_every_iteration_num: int = 5,
        fallback_optimization: Optimizer | None = None,
        super_iteration_tol: float = 1e-7,
        seed: int | None = None,
    ):
        super().__init__(measure, optimization_map)
        self._super_iteration_every = super_iteration_every
        self._max_super_iteration_num = max_super_iteration_num
        self._num_print_every = print_every_iteration_num
        self.callback = self._callback_function
        self._old_parameters_dict = {}
        self._old_opt_idxs_dict = {}
        self._fallback_optimization = fallback_optimization
        self._super_iteration_tol = super_iteration_tol
        self._seed = seed

    def _dcrab_super_iteration(self) -> None:
        optimizables = list(self._optimization_map.get_optimizables())
        relevant_optimizables = []

        dcrab_envs = []
        for opt in optimizables:
            if isinstance(opt, DCRABEnvelope):
                dcrab_envs.append(opt)
                relevant_optimizables.append(opt)
            else:
                try:
                    dcrab_envs.extend(opt.get_envelopes())  # type:ignore
                    relevant_optimizables.append(opt)  # type:ignore
                except Exception as e:
                    raise ConfigurationException(
                        "Optimizable is not a `DCRABEnvelope` nor does it have a `get_envelope` method.\n"
                        + f"Raised exception {e}"
                    )

        # Add new parameters to the dCRAB envelope
        for n, env in enumerate(dcrab_envs):
            try:
                if self._seed is not None:
                    seed = (n + 1) * self._super_iteration_num * self._seed
                else:
                    seed = None
                env.add_new_components(seed=seed)
            except Exception as e:
                raise ConfigurationException(f"Non-`DCRABEnvelope` encountered. \n Raised exception {e}")

        # Add new parameters to optmap
        new_coeffs_freqs_and_phases = [env.get_coefficients_frequencies_and_phases() for env in dcrab_envs]
        for env, coeffs_freqs_phases in zip(relevant_optimizables, new_coeffs_freqs_and_phases):
            params = [env.get_parameters()[0]] + coeffs_freqs_phases  # Add pulse amplitude to optimization
            if isinstance(env, DRAGMixer):
                params.append(env.get_parameters()[-1])  # Add pulse delta to optimization if it is a DRAG tone.
            self._optimization_map.add(env, params)

        # Get all parameters and update scales for optimization
        params = self._optimization_map.get_all_parameters()
        self._scales = jnp.array([p.get_scale() for p in params]).flatten()
        print(f"* Current no. of parameters = {len(params)}")

        # Restart the optimization process
        self._build_optimizable_index_list()
        self._optimization_map.register_params_with_optimizables()

        init = []
        for qty in self._optimization_map.get_all_parameters():
            init.append(qty.get_reduced_value())

        self._minimize_infidelity(init)

    def _callback_function(self, intermediate_result):
        if self._num_iteration % self._num_print_every == 0:
            print(f"Iteration number = {self._num_iteration} \t  Infidelity  = {intermediate_result.fun:.3e}")

        self._num_iteration += 1
        self._super_iteration_since += 1

        if intermediate_result.fun < self.best_fid:
            self.best_fid = intermediate_result.fun
            self.best_params = intermediate_result.x

        if self._super_iteration_num >= self._max_super_iteration_num:
            raise StopIteration("Maximum number of super iterations performed. Stopping optimization.")

        if np.abs(intermediate_result.fun - self._previous_fid) < self._super_iteration_tol:
            self._super_iteration_num += 1

            print("\n")
            print(f"==== Decrease in infidelity less than {self._super_iteration_tol} ====")
            print(f"==== Starting super-iteration {self._super_iteration_num} ====")
            print(f"* Current lowest infidelity = {self.best_fid: .3e}")

            current_params = self._optimization_map.get_all_parameters()
            self._old_parameters_dict[len(current_params)] = current_params
            self._old_opt_idxs_dict[len(current_params)] = self._opt_idxs

            # Start the dCRAB super-iteration
            self._super_iteration_since = 0
            self._dcrab_super_iteration()

        self._previous_fid = intermediate_result.fun

        if self._super_iteration_since >= self._super_iteration_every:
            self._super_iteration_num += 1

            print("\n")
            print("==== Max iteration before a super-iteration reached ====")
            print(f"==== Starting super-iteration {self._super_iteration_num} ====")
            print(f"*** Current lowest infidelity = {self.best_fid: .3f} ***")

            current_params = self._optimization_map.get_all_parameters()
            self._old_parameters_dict[len(current_params)] = current_params
            self._old_opt_idxs_dict[len(current_params)] = self._opt_idxs

            # Start the dCRAB super-iteration
            self._super_iteration_since = 0
            self._dcrab_super_iteration()

    def set_parameters(self, values):
        """Update the parameter values.

        This method is derived from the `ScipyoptimizerGradient` class and designed
        to catch cases involving mismatch in dimension of parameters.

        Since, in dCRAB, new parameters are added in each super-iteration,
        the previous best result may be one with fewer parameters.
        In that case, all subsequent parameters are set to their minimum value (by setting the reduced value to -1).

        Parameters
        ----------
        values: Array
            Parameter values for the update.

        Returns
        -------
        Array
            Returns the inverse of the fidelity.

        """
        log = []
        params = self._optimization_map.get_all_parameters()

        # Check for edge case when best result has fewer parameters than the current iteration.
        if len(params) != len(values):
            num_values = len(values)
            warnings.warn(
                "Stopping the optimization or backtracking to previous best fidelity."
                + f" Going to step with {num_values} parameters."
            )
            if num_values in self._old_parameters_dict:
                ids = [id(i) for i in self._old_parameters_dict[num_values]]
                j = 0
                for ii in range(len(params)):
                    if id(params[ii]) in ids:
                        if j >= num_values:
                            raise Exception("Possible duplicate entries in the optmap.")
                        params[ii].set_reduced_value(np.array([values[j]]))
                        j += 1
                    else:
                        params[ii].set_reduced_value(np.array([0]))
                    log.append(params[ii])
            else:
                raise ConfigurationException(
                    f"Got values from optimization than does not fit in optimization map. \n \
                    No. of values from optimization = {num_values} \
                    and no. of objects in optimization map \
                    (in the different super-iterations) = {list(self._old_parameters_dict.keys())}."
                )
        else:
            for index, val in enumerate(np.split(values, self._opt_idxs[:-1])):
                params[index].set_reduced_value(val)
                log.append(params[index])
        return log

    def _set_parameters_and_measure(self, values) -> float:
        """Update the parameter values and return measurement result.

        Parameters
        ----------
        values: Array
            Parameter values for the update.

        Returns
        -------
        Array
            Returns the inverse of the fidelity.

        """
        log = self.set_parameters(values)

        if isinstance(self._measure, Differentiable):
            fun, grad = self._measure.get_value_and_gradient(self._times)
            self._grad_cache = grad

        infid = 1.0 - fun
        if self._logger:
            self._logger.log(log, float(infid))
        return float(1 - fun)

    def _minimize_infidelity(self, init):
        result = minimize(
            fun=self._set_parameters_and_measure,
            jac=self._lookup_jac,
            x0=jnp.concatenate(init).flatten(),
            bounds=[(-1, 1)] * self._opt_idxs[-1],
            method=self._method,
            options=self._options,
            callback=self._callback,
        )
        return result

    def optimize(self, times: Array | float) -> OptimizationResult:
        """Optimize via the Scipy optimizer gradient model.

        Performs the actual optimization.

        *Note - If input `times` is a float, then the start time of propagation is implicitly assumed to be zero.
        For an array of times, the first time point is the start time.*

        Returns
        -------
        OptimizationResult
            The result of the optimization.

        """
        self._times = jnp.array([0.0, times]) if isinstance(times, float) else times

        if self._logger:
            self._logger.start()

        self._build_optimizable_index_list()
        self._optimization_map.register_params_with_optimizables()

        init = []
        for qty in self._optimization_map.get_all_parameters():
            init.append(qty.get_reduced_value())

        self.best_params = list(jnp.concatenate(init).flatten())

        try:
            result = self._minimize_infidelity(init)
        except Exception as e:
            if "_lbfgsb._lbfgsb.setulb: failed to create array from the 7th" + " argument `g`" in str(e):
                raise IncompatibleOptimizationMap(
                    "Number of quantities in optMap differ from number of" + f" gradients computed. \n {e}"
                )
            else:
                raise e

        # Set the parameters to the best params at the end of optimization
        print("Setting parameters to the best values.")
        self.set_parameters(self.best_params)

        if self._fallback_optimization is not None:
            print("Performing fallback optimization")
            self._fallback_optimization.optimize(self._times)

        if self._logger:
            self._logger.stop(str(result))

        return OptimizationResult(
            status=(OptimizationResult.STATUS_SUCCESS if result.success else OptimizationResult.STATUS_FAILED),
            value=self.best_fid,
            iterations=result.nfev,
            message=result.message,
            raw_result=result,
        )



import numpy as np
import pytest

from paraqeet.signal.waveform import LocalOscillator


@pytest.fixture
def local_oscillator() -> LocalOscillator:
    return LocalOscillator()


@pytest.fixture
def random_time_vector():
    return np.linspace(0.0, np.random.randint(1, 10) * np.random.rand(), np.random.randint(1, 10))


def test_getters_and_setters(local_oscillator, random_quantity):
    frequency = random_quantity(1, "Hz")
    local_oscillator.frequency = frequency
    assert local_oscillator.frequency == frequency

    assert local_oscillator.get_parameters() is not None
    assert len(local_oscillator.get_parameters()) >= 0
    assert frequency in local_oscillator.get_parameters()


def test_output_shapes(local_oscillator, random_time_vector):
    output = local_oscillator.get_value(random_time_vector)
    assert output.shape == (len(random_time_vector),)

    local_oscillator.set_optimizable_parameters(local_oscillator.get_parameters())
    _, gradient = local_oscillator.get_value_and_gradient(random_time_vector)
    assert gradient.shape == (len(random_time_vector), len(local_oscillator.get_parameters()))

    time_gradient = local_oscillator.get_time_gradient(random_time_vector)
    assert time_gradient.shape == (len(random_time_vector),)



"""Class definition for the Optimizable Map model."""

from collections.abc import Callable

from paraqeet.exceptions import SerializationException
from paraqeet.optimizable import Optimizable
from paraqeet.quantity import Quantity


class OptimizationMap:
    """Optimization parameter map utility class.

    Utility class that collects all parameters that shall be considered during
    optimization and associates them with the corresponding Optimizable
    interface. With this class, Quantities can be traced back to the
    Optimizable to which they belong. Before optimization, an instance of this
    class needs to be filled and passed to the optimizer.

    """

    _optimizable_to_parameter_map: dict[Optimizable, list[Quantity]]

    def __init__(self):
        self._optimizable_to_parameter_map = {}

    def __repr__(self):
        """Magic method for human readable representation."""
        return self.__str__()

    def __str__(self):
        """Human readable representation of the parameters set to optimize."""
        om_str = ""
        for key, val in self._optimizable_to_parameter_map.items():
            om_str += f"==== {key} ====\n"
            om_str += str(val)
            om_str += "\n\n"
        return om_str

    def add(
        self,
        optimizable: Optimizable,
        optimizable_quantities: Quantity | list[Quantity] | None = None,
    ):
        """Add an optimizable object and a list of its quantities to the map.

        The list contains all parameters of the optimizable object that shall
        be considered during the optimization. If the list is empty, all
        parameters of the class will be used. If the object was already added,
        the list of quantities will be overwritten.

        Parameters
        ----------
        optimizable: Optimizable
            Input Optimizable object for adding to the map.
        optimizable_quantities: Quantity | list[Quantity] | None = None
            List of all parameters of the optimizable object considered for
            optimization.

        """
        if optimizable_quantities is not None:
            params: list[Quantity] = (
                [optimizable_quantities] if isinstance(optimizable_quantities, Quantity) else optimizable_quantities
            )
        else:
            params = optimizable.get_parameters()
        self._optimizable_to_parameter_map[optimizable] = params
        if len(self._optimizable_to_parameter_map[optimizable]) < 1:
            self._optimizable_to_parameter_map.pop(optimizable)

    def append(
        self,
        optimizable: Optimizable,
        optimizable_quantities: Quantity | list[Quantity] | None = None,
    ):
        """Append an optimizable object and a list of its quantities to the map.

        This method is similar to the `add` method, but instead of overwriting the
        existing entries, this appends the specified list of quantities to the
        already existing quantities.

        Parameters
        ----------
        optimizable: optimizable
            Input optimizable object for adding to the map.
        optimizable_quantities: Quantity | list[Quantity] | None = None
            List of all parameters of the optimizable object considered for
            optimization.

        """
        if optimizable_quantities is not None:
            params: list[Quantity] = (
                [optimizable_quantities] if isinstance(optimizable_quantities, Quantity) else optimizable_quantities
            )
        else:
            params = optimizable.get_parameters()

        if optimizable in self._optimizable_to_parameter_map:
            self._optimizable_to_parameter_map[optimizable].extend(params)
        else:
            self._optimizable_to_parameter_map[optimizable] = params
        if len(self._optimizable_to_parameter_map[optimizable]) < 1:
            self._optimizable_to_parameter_map.pop(optimizable)

    def remove(self, optimizable: Optimizable, params: Quantity | list[Quantity] | None = None):
        """Remove the given optimizable or parameter(s) from the optimization map.

        If params is None, it removes the optimizable from the optimization map.
        Else it only removes the specific parameter from the optimization map.

        Parameters
        ----------
        optimizable: optimizable
            optimizable to be removed.
        params: Quantity | list[Quantity] | None.
            Parameter(s) to be removed from the optimization map. If None removes the optimizable.

        """
        try:
            if params is None:
                self._optimizable_to_parameter_map.pop(optimizable)
            else:
                parameters_list: list[Quantity] = params if isinstance(params, list) else [params]
                self.filter_parameters(lambda quantity: quantity not in parameters_list)

        # removed the bare except catch.
        except Exception as e:
            raise Exception(e)

    def replace(
        self,
        optimizable: Optimizable,
        old_parameters: Quantity | list[Quantity],
        new_parameters: Quantity | list[Quantity],
    ) -> None:
        """Perform an in-place substitution of the old and new parameters.
        This helps to keeps the ordering of parameters the same while replacing parameters.
        """
        old_parameters_list = old_parameters if isinstance(old_parameters, list) else [old_parameters]
        new_parameters_list = new_parameters if isinstance(new_parameters, list) else [new_parameters]

        for old_param, new_param in zip(old_parameters_list, new_parameters_list):
            for i, param in enumerate(self._optimizable_to_parameter_map[optimizable]):
                if id(param) == id(old_param):
                    self._optimizable_to_parameter_map[optimizable][i] = new_param

    def get_optimizables(self) -> set[Optimizable]:
        """Return all optimizable objects that were added to this map.

        Returns
        -------
        set[Optimizable]
            Set of all optimizable objects from the map.

        """
        return set(self._optimizable_to_parameter_map.keys())

    def get_parameters(self, optimizable: Optimizable) -> list[Quantity] | None:
        """Return all quantities associated with the given parameter.

        Parameters
        ----------
        optimizable: Optimizable
            Input optimizable object.

        Returns
        -------
        list[Quantity] | None
            List of parameters or None (if the optimizable has not been
            added yet).

        """
        return self._optimizable_to_parameter_map[optimizable]

    def get_all_parameters(self) -> list[Quantity]:
        """Return all parameters that were added to the system map.

        Returns
        -------
        List[Quantity]
            All parameters that were added to the map.

        """
        quantities = []
        for params in self._optimizable_to_parameter_map.values():
            quantities.extend(params)
        return quantities

    def register_params_with_optimizables(self) -> None:
        """Register optimizable parameters with the system.

        Utility function that synchronises the list of parameters with
        each optimizable class. This needs to be called by the optimizer
        before gradient based optimization to tell the layers which gradients
        to compute.

        """
        for optimizable, params in self._optimizable_to_parameter_map.items():
            optimizable.set_optimizable_parameters(params)
            optimizable.set_all_optimizable_parameters(self.get_all_parameters())

    def filter_parameters(self, filter_function: Callable) -> None:
        """Filter parameters using filter function.

        Updates the list of parameters for all Optimizables in this map using
        a filter function. Only parameters for which the filter function returns
        true will remain in this map.

        Parameters
        ----------
        filter_function : Callable
            Filter function that maps quantities to boolean values.

        """
        for key in self._optimizable_to_parameter_map.keys():
            filtered = filter(filter_function, self._optimizable_to_parameter_map[key])
            self._optimizable_to_parameter_map[key] = list(filtered)
        self._optimizable_to_parameter_map = dict(
            (k, v) for k, v in self._optimizable_to_parameter_map.items() if len(v) > 0
        )

    def filter_by_name(self, name: str):
        """Filter parameters by name of parameter.

        Parameters
        ----------
        name : str
            Name of parameter to be filtered with.

        """
        return self.filter_parameters(lambda quantity: quantity.get_name() == name)

    def remove_by_name(self, name: str):
        """Remove parameters by name of parameter.

        Parameters
        ----------
        name : str
            Name of parameter to be filtered with.

        """
        return self.filter_parameters(lambda quantity: name not in quantity.get_name())

    def to_dict(self) -> dict:
        """Create a dictionary represenation of the optimization map.

        Creates a dictionary that contains the values of all quantities that are being optimized, sorted by the
        Optimizable instances to which they belong. The returned dictionary is meant for export using the serialisation
        package. It uses the names of Optimizables and Quantities and assumes that those are unique and not None. The
        format of the dict will be

        .. code-block:: python

            "optimizable name": {
                "quantity name": {
                "unit": string,
                "shape": tuple[int, ...],
                "twoPi": bool,
                "value': Array,
                "min": Array,
                "max": Array
                }
            }

        where the innermost part is generated by Quantity's toDict function.


        Returns
        -------
        dict
            all optimized quantities in an exportable format

        Raises
        ------
        SerialisationException
            If the name of any Optimizable or Quantity is None or not unique.

        """
        data = dict()
        for optimizable, quantities in self._optimizable_to_parameter_map.items():
            # Check that the optimizable's name is valid
            if len((optimizable.name or "").strip()) == 0 or optimizable.name in data:
                raise SerializationException("Optimizable does not have a name or the name is not unique.")

            # Check that the quantities' names are valid
            quantity_names = [(q.get_name() or "").strip() for q in quantities]
            non_empty_quantity_names = list(filter(lambda name: len(name) > 0, quantity_names))
            if len(quantities) != len(set(non_empty_quantity_names)):
                raise SerializationException(
                    f"Quantities in {optimizable.name} have empty or non-unique names within the optimizable."
                )

            data[optimizable.name] = {q.get_name(): q.to_dict() for q in quantities}
        return data

    def from_dict(self, data: dict) -> None:
        """
        Restores the values of all optimized quantities that are in the dictionary. The format of the dictionary needs
        to be in the same format as generated by the toDict function.

        Parameters
        ----------
        data: dict
            All quantities that should be restored.

        Raises
        ------
        SerialisationException
            If the dict contains an Optimizable or a Quantity that does not exist in this optimization map.
        """
        optimizables_for_name = {opt.name: opt for opt in self._optimizable_to_parameter_map.keys()}
        for optimizable_name, values in data.items():
            if optimizable_name not in optimizables_for_name:
                raise SerializationException(
                    f'An optimizable with the name "{optimizable_name}" does not exist in the optimization map.'
                )
            optimizable = optimizables_for_name[optimizable_name]

            quantities_for_name = {q.get_name(): q for q in optimizable.get_parameters()}
            for quantity_name, quantity_values in values.items():
                if quantity_name not in quantities_for_name:
                    raise SerializationException(f'Quantity "{quantity_name}" does not exist in {optimizable_name}.')
                quantities_for_name[quantity_name].from_dict(quantity_values)



"""Class definition of the Resonator Hamiltonian model."""

import jax
import jax.numpy as jnp

from paraqeet.exceptions import ConfigurationException
from paraqeet.model.differentiable_hamiltonian import DifferentiableHamiltonian
from paraqeet.model.drive import Drive
from paraqeet.quantity import Array, Quantity

jax.config.update("jax_enable_x64", True)

# TODO: Why is a resonator a Hamiltonian? Isn't a resonator a quantum system which has a Hamiltonian
#  as a property? What does “derivative of a resonator” mean physically?
class Resonator(DifferentiableHamiltonian):
    """Hamiltonian of a harmonic oscillator.

    The only optimizable parameter is the frequency.

    Parameters
    ----------
    dimension : int
        Dimension of the harmonic oscillator.
    frequency : Quantity
        Frequency of the harmonic oscillator.
    drives : list[Drive], optional
        List of time-dependent drives of the subsystem.

    """

    _dimension: int
    _frequency: Quantity
    _annihilation_op: Array
    _num_op: Array
    _t1: Quantity | None
    _temp: Quantity | None
    _t2star: Quantity | None

    # TODO: we should think about the composition here instead of inheritance from DifferentiableHamiltonian.
    def __init__(
        self,
        dimension: int,
        frequency: Quantity,
        drives: list[Drive] | None = None,
        t1: Quantity | None = None,
        temp: Quantity | None = None,
        t2star: Quantity | None = None,
    ):
        super().__init__(drives=drives)
        self._dimension = dimension
        self._frequency = frequency
        self._annihilation_op = jnp.sqrt(jnp.diag(jnp.arange(1, dimension, dtype=jnp.float64), k=1))
        self._num_op = self._annihilation_op.T @ self._annihilation_op
        self.t1 = t1
        self.temp = temp
        self.t2star = t2star

    def dimension(self):
        """Get the dimension of the resonator."""
        return self._dimension

    @property
    def frequency(self) -> Quantity:
        """Get the frequency of the resonator."""
        return self._frequency

    @frequency.setter
    def frequency(self, frequency: Quantity) -> None:
        """Set the frequency of the resonator."""
        self._frequency = frequency

    @property
    def t1(self) -> Quantity | None:
        """Get the t1 of the resonator."""
        return self._t1

    @t1.setter
    def t1(self, t1: Quantity | None) -> None:
        """Set the t1 of the resonator."""
        self._t1 = t1

    @property
    def temp(self) -> Quantity | None:
        """Get the temp of the resonator."""
        return self._temp

    @temp.setter
    def temp(self, temp: Quantity | None) -> None:
        """Set the temp of the resonator."""
        self._temp = temp

    @property
    def t2star(self) -> Quantity | None:
        """Get the t2star of the resonator."""
        return self._t2star

    @t2star.setter
    def t2star(self, t2star: Quantity | None) -> None:
        """Set the t2star of the resonator."""
        self._t2star = t2star

    def get_parameters(self) -> list[Quantity]:
        """Get parameters of the model.

        Returns
        -------
        List[Quantity]
            Returns the list of parameters of the system.

        """
        return self._get_drive_parameters() + [self._frequency]

    def get_value_at_timestep(self, timestep: float) -> Array:
        """Get the drive matrix.

        Parameters
        ----------
        timestep : float
            One time stamp.

        Returns
        -------
        Array
            The drive matrix at a single timestamp.

        """
        H = self._frequency.get_value() * self._num_op
        return H + self._get_drive_matrix_at_timestep(self._annihilation_op, timestep)

    def get_gradient_at_timestep(self, time: float) -> Array:
        """Get the gradient of the drive.

        Parameters
        ----------
        times : float
            One time stamp.

        Returns
        -------
        Array
            Returns the gradients of the drive.

        """
        # Fetch the gradient of the drive
        derivatives = self._get_drive_gradients_at_timestep(self._annihilation_op, time)

        # Combine with the derivative wrt the frequency
        if self._is_optimized(self._frequency):
            grad = self._num_op.reshape((1,) + self._num_op.shape)
            derivatives = jnp.append(derivatives, grad, axis=0)
        return derivatives

    def get_decay_rates(self) -> list[Array]:
        """Return decay rate for T1, T2star and Temp respectively."""
        if (self.t1 is None) or (self.t2star is None) or (self.temp is None):
            raise ConfigurationException("Specify values of T1, T2star and Temp for Open system simulations.")

        gamma = 1 / self.t1.get_value()
        gamma_t2star = 0.5 / self.t2star.get_value()

        hbar_over_kb = 7.638232582257738e-12
        beta = hbar_over_kb / (self.temp.get_value())
        nbar = jnp.exp(-beta * self.frequency.get_value())
        gamma_temp = gamma * nbar
        gamma_t1 = gamma * (nbar + 1)
        return [gamma_t1, gamma_temp, gamma_t2star]

    def get_collapseops(self) -> list[tuple[Array, Array]]:
        """
        Return a list tuples of decay rates and collapse operators for each subsystem.

        Return
        ------
        list[tuple[Array, Array]]
            List of collapse operators
        """
        gamma_t1, gamma_temp, gamma_t2star = self.get_decay_rates()
        col_t1 = self._annihilation_op
        col_temp = self._annihilation_op.T
        col_t2star = 2 * self._num_op
        return [(gamma_t1, col_t1), (gamma_temp, col_temp), (gamma_t2star, col_t2star)]



"""Test the Qubit model."""

import numpy as np
import pytest

from paraqeet.differentiable import Differentiable
from paraqeet.exceptions import ConfigurationException
from paraqeet.model.drive_operator import DriveOperator
from paraqeet.model.qubit import Qubit
from paraqeet.quantity import Quantity
from paraqeet.signal.envelopes import FlatTopGaussianEnvelope
from paraqeet.signal.iq_mixer import IQMixer

FREQ = 4.8e9 * 2 * np.pi
LEN_SIG = 1001


@pytest.fixture
def time_samples():
    """Generate time samples from the given signal length.

    The given signal length `LEN_SIG` is a module level global variable.

    """
    return np.linspace(0, 10e-9, LEN_SIG)


@pytest.fixture
def tone():
    """Return a object for sinusoidal tone generator with error envelopes."""
    return FlatTopGaussianEnvelope()


@pytest.fixture
def gen(tone):
    """Return a sinusoidal tone generator object."""
    gen = IQMixer(envelopes=[tone])
    return gen


@pytest.fixture
def ham(gen):
    """Return a qubit."""
    drive = DriveOperator(gen, is_longitudinal=False)
    return Qubit(Quantity(FREQ, 0.8 * FREQ, 1.2 * FREQ), drives=[drive])


def test_get_value(ham, time_samples):
    """Test the getMatrix method."""
    hams = ham.get_value(time_samples)
    assert hams.shape == time_samples.shape + (2, 2)


def test_gradient(gen, ham, time_samples):
    """Test the gradients of the system.

    The Hamiltonian should have all derivatives of the drive
    plus the derivative w.r.t. the qubit frequency.

    """
    if not isinstance(ham, Differentiable):
        return

    _, grads = gen.get_value_and_gradient(time_samples)
    ham.set_optimizable_parameters(ham.get_parameters())
    # TODO: fix error related to the length of time_samples-array
    _, ham_grads = ham.get_value_and_gradient(time_samples)
    assert ham_grads.shape == (len(time_samples), grads.shape[1] + 1, 2, 2)

    ham.set_optimizable_parameters([ham.frequency])
    _, ham_grads = ham.get_value_and_gradient(time_samples)
    assert ham_grads.shape == (len(time_samples), 1, 2, 2)


def test_setters_and_getters(ham, random_quantity):
    frequency = random_quantity(1)
    ham.frequency = frequency
    assert ham.frequency == frequency


def test_needs_parameters_for_decay_rates(ham):
    for _ in range(10):
        t1 = Quantity(1.0, 0.0, 2.0) if np.random.random() < 0.8 else None
        t2star = Quantity(1.0, 0.0, 2.0) if np.random.random() < 0.8 else None
        temp = Quantity(1.0, 0.0, 2.0) if np.random.random() < 0.8 else None
        ham.t1 = t1
        ham.t2star = t2star
        ham.temp = temp
        if t1 is not None and t2star is not None and temp is not None:
            # Valid parameters should work
            ham.get_collapseops()
        else:
            # Invalid parameters should raise an exception
            with pytest.raises(ConfigurationException):
                ham.get_collapseops()



"""Class definition of the Generator model."""

from abc import abstractmethod

from paraqeet.differentiable import Differentiable
from paraqeet.optimizable import Optimizable
from paraqeet.quantity import Array


class Generator(Optimizable, Differentiable):
    """Signal generation stack.

    Contrary to most quantum simulators, paraqeet includes a detailed simulation
    of the control stack. Each component in the stack and its functions are
    simulated individually and combined here.

    Example: A local oscillator and arbitrary waveform generator signal
    are put through via a mixer device to produce an effective modulated signal.

    """

    @abstractmethod
    def get_value(self, times: Array | float) -> Array:
        """Return array with scalar signal value for each time step.

        Parameters
        ----------
        times: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the scalar signal vector.

        """
        pass

    @abstractmethod
    def get_gradient_at_timestep(self, time: float) -> Array:
        """Return array with the gradient of the signal value for one time step.

        The result has the shape (p,) where 'p' is the parameter index.

        Parameters
        ----------
        time: Array
            One time stamp.

        Returns
        -------
        Array

        """
        pass



# TODO: implement abstract class providing a method get_gradient

# All signal generators are differentiable?
# Is a default implementation possible? Not yet
#  Subclasses of Measurement lacking the implementation of the gradient calculation are
# NOT Differentiables? Yes

from abc import ABC, abstractmethod

from paraqeet.quantity import Array


class Differentiable(ABC):
    """An abstract class for differentiable models.

    Subclasses must implement the value_and_gradient() method which would
    return the gradient of the model.
    """

    @abstractmethod
    def get_value_and_gradient(self, times: Array) -> tuple[Array, Array] | tuple[float, Array]:
        """Calculate the gradient of the model.

        Returns
        -------
        tuple[Array, Array] | tuple[float, Array]
            The value and the gradient of the model.

        """
        pass



from abc import abstractmethod

from jax import vmap

from paraqeet.differentiable import Differentiable
from paraqeet.model.hamiltonian import Hamiltonian
from paraqeet.quantity import Array


class DifferentiableHamiltonian(Differentiable, Hamiltonian):
    """
    A Hamiltonian that also satisfies the differentiable interface. It requires a gradient function that works on a
    per timestep basis and provides a vectorized mapping to evaluate multiple time values.
    """

    @abstractmethod
    def get_gradient_at_timestep(self, time: float) -> Array:
        """Compute the gradient of this Hamiltonian wrt to parameters for a single timestep."""
        pass

    def get_value_and_gradient(self, times: Array) -> tuple[Array, Array]:
        """Compute value and gradient for given timesteps. The gradient call uses vmap over at_timestep methods."""
        # ignoring mypy due to vmap
        return self.get_value(times), vmap(self.get_gradient_at_timestep)(times)  # type: ignore



"""paraqeet: A quantum optimal control toolkit with simple parameter management."""

from paraqeet.optimization_map import OptimizationMap as OptimizationMap
from paraqeet.quantity import Array as Array
from paraqeet.quantity import Quantity as Quantity



"""Test the Scipy piecewise exponentiation solver."""

import numpy as np
import pytest

from paraqeet.propagation.scipy_expm import ScipyExpm
from tests.model.dummy_model import DummyEquationsOfMotion
from tests.model.empty_hamiltonian import EmptyHamiltonian
from tests.propagation.test_common_propagation import needs_initial_state


@pytest.fixture
def expm():
    """Return a Scipy piecewise exponentitation solver generating method."""

    def _method(dimension, resolution):
        return ScipyExpm(DummyEquationsOfMotion(EmptyHamiltonian(dimension)), resolution=resolution)

    return _method


def test_parameters(expm):
    """Test parameters from the propagation."""
    propagation = expm(dimension=np.random.randint(10), resolution=3)
    assert propagation.get_parameters() == []


def test_resolution(expm):
    """Test the resolution after propagation."""
    for i in range(10):
        propagation = expm(dimension=np.random.randint(2, 100), resolution=3)
        resolution = np.random.randint(1, 1000)
        propagation.resolution = resolution
        assert propagation.resolution == resolution


def test_state_dimension_vector(random_state, expm, ts):
    """Test the dimension and the norm of state vectors.

    The dimension and norm of state vectors should be the
    same after propagation.

    """
    for _ in range(10):
        dim = np.random.randint(2, 30)
        state = random_state(dim)
        propagation = expm(dim, resolution=3)
        propagation.set_initial_state(state)
        propagated_states = propagation.propagate(ts)
        assert propagated_states.shape[0] == len(ts)
        assert propagated_states.shape[1:] == state.shape + (1,)


def test_state_dimension_rect_matrix(random_matrix, expm, ts):
    """Test the state matrix after propagation."""
    for _ in range(10):
        basis = np.random.randint(2, 10)
        dim = basis + np.random.randint(1, 3)
        state = random_matrix(dim, basis)  # rect matrix with dim>basis
        propagation = expm(dim, resolution=3)
        propagation.set_initial_state(state)
        propagated_states = propagation.propagate(ts)
        assert propagated_states.shape[0] == len(ts)
        assert propagated_states.shape[1:] == state.shape


def test_state_dimension_square_matrix(expm, ts):
    """Test for a batch of initial states for propagation."""
    for _ in range(5):
        basis = np.random.randint(2, 10)
        dim = basis + np.random.randint(1, 3)
        state = np.eye(dim, dtype=np.complex128)
        propagation = expm(dim, resolution=3)
        propagation.set_initial_state(state)
        propagated_states = propagation.propagate(ts)
        assert propagated_states.shape[0] == len(ts)
        assert propagated_states.shape[1:] == state.shape


def test_state_dimension_matrix_open(random_matrix, expm, ts):
    """Test the state matrix after propagation."""
    for _ in range(10):
        dim = np.random.randint(2, 10)
        state = random_matrix(dim, dim)
        propagation = expm(dim**2, resolution=3)
        propagation.is_open = True
        propagation.set_initial_state(state)
        propagated_states = propagation.propagate(ts)
        assert propagated_states.shape[0] == len(ts)
        assert propagated_states.shape[1:] == state.shape


def test_state_dimension_square_matrix_open(expm, ts):
    """Test for a batch of initial states for propagation."""
    for _ in range(5):
        dim = np.random.randint(2, 10)
        state = np.eye(dim, dtype=np.complex128)
        propagation = expm(dim**2, resolution=3)
        propagation.is_open = True
        propagation.set_initial_state(state)
        propagated_states = propagation.propagate(ts)
        assert propagated_states.shape[0] == len(ts)
        assert propagated_states.shape[1:] == state.shape


def test_initial_state(model):
    """Test whether the initial state is set.

    Raises
    ------
    ConfigurationException
        If the initial state is not set.

    """
    dim = np.random.randint(2, 10)
    m = model(dim)
    propagation = ScipyExpm(model=m, resolution=3)
    needs_initial_state(propagation, dim)


def test_construct_times(model):
    """Test the construction times for the model."""
    res = 100e9

    propagation = ScipyExpm(model=model, resolution=res)

    # Test if times array is constructed correctly for a 1ns list
    t_start = 1e-9
    t_final = 2e-9

    time = np.array([t_start, t_final])
    steps = int(np.ceil((t_final - t_start) * res))
    full_times = np.linspace(t_start, t_final, steps, endpoint=False)
    times, dt = propagation._construct_times(time, 1)

    assert np.allclose(times, full_times)
    assert np.isclose(dt, 1 / res)

    # Test if times array is constructed correctly if steps < 2
    t_start = 1e-9
    t_final = t_start + (1 / res)

    time = np.array([t_start, t_final])
    steps = int(np.ceil((t_final - t_start) * res))
    full_times = np.linspace(t_start, t_final, steps, endpoint=False)
    times, dt = propagation._construct_times(time, 1)

    assert np.allclose(time, full_times)
    assert np.isclose(dt, 1 / res)



"""Class definition of a coupling optimizable model."""

import jax
import jax.numpy as jnp

from paraqeet.model.differentiable_hamiltonian import DifferentiableHamiltonian
from paraqeet.optimizable import Optimizable
from paraqeet.quantity import Array, Quantity

jax.config.update("jax_enable_x64", True)


# TODO: Make this a 2-system coupling class, finish in notebooks
class TwoBodyCoupling(Optimizable):
    """Create a coupling optimizable model.

    Represents the coupling of two or more subsystems in a composite
    Hamiltonian. This class implements longitudinal and transversal
    coupling with a constant scalar coefficient.
    The coefficient is the only optimizable parameter.
    Subclasses can alter the behavior by overriding the `get_couplings` function.

    Parameters
    ----------
    subsystems : list[Hamiltonian]
        The coupled subsystems.
    coefficient : Quantity
        Either a constant coefficient as float or a callable that returns the
        coefficient for a given time.
    is_longitudinal : bool
        Whether the coupling is longitudinal or transversal.
    use_rwa : bool, optional
        If the transversal coupling should use the rotating-wave approximation
        or should include double excitation terms.
    """

    _subsystem_A: DifferentiableHamiltonian
    _subsystem_B: DifferentiableHamiltonian
    _coefficient: Quantity
    _total_dims: int
    _is_longitudinal: bool
    _use_rwa: bool

    def __init__(
        self,
        subsystem_A: DifferentiableHamiltonian,
        subsystem_B: DifferentiableHamiltonian,
        coefficient: Quantity,
        is_longitudinal: bool,
        use_rwa: bool = False,
    ):
        self._subsystem_A = subsystem_A
        self._subsystem_B = subsystem_B
        self._coefficient = coefficient
        self._is_longitudinal = is_longitudinal
        self._use_rwa = use_rwa
        self._total_dims = subsystem_A.dimension() * subsystem_B.dimension()
        self._dims = [subsystem_A.dimension(), subsystem_B.dimension()]
        self._annihilation_op = [jnp.sqrt(jnp.diag(jnp.arange(1, dim, dtype=jnp.float64), k=1)) for dim in self._dims]

    def get_parameters(self) -> list[Quantity]:
        """Collect parameters from all subsystems and couplings.

        Parameters
        ----------
        list[Quantity]
            Returns the list of parameters of the system.

        """
        return [self._coefficient]

    @property
    def subsystem_A(self) -> DifferentiableHamiltonian:
        """Return subsystem A that is coupled by this term.

        Returns
        -------
        list[Hamiltonian]
            List of subsystems.

        """
        return self._subsystem_A

    @property
    def subsystem_B(self) -> DifferentiableHamiltonian:
        """Return subsystem A that is coupled by this term.

        Returns
        -------
        list[Hamiltonian]
            List of subsystems.

        """
        return self._subsystem_B

    @property
    def subsystems(self) -> list[DifferentiableHamiltonian]:
        """Return the subsystems as a list to be compatible with other couplings with potentially more subsystems."""
        return [self.subsystem_A, self.subsystem_B]

    def get_couplings(self) -> list[list[Array]]:
        """Return the matrix representation of a time-indepedent coupling.

        Returns
        -------
        list[Array]
            The outer list are the coupling terms. The inner list represents
            the subsystems. The matrices (Array) have the same shape as the
            subsystem's Hamiltonian.get_value: (t,n,n) with t the time and n
            the subsystem dimension.

        """
        matrices = self._coupling_operators()
        for i in range(len(matrices)):  # iterating over terms
            matrices[i][0] *= self._coefficient.get_value()
        return matrices

    def get_coupling_gradients(self) -> list[list[list[Array]]]:
        """Return the gradients of the coupling terms.

        Returns
        -------
        list[Array]
            The outer list represents the gradients with respect to all
            optimized parameters.
            The rest is in the same shape as the result of get_matrices.

        """
        coup_ops = self._coupling_operators()
        if self._is_optimized(self._coefficient):
            grads = [coup_ops]
        else:
            grads = [[[jnp.empty((0, 0)) for _ in sub] for sub in coup_ops]]
        return grads

    def _coupling_operators(self) -> list[list[Array]]:
        """Return coupling operators.

        Returns the operators of the longitudinal or transversal coupling
        without coefficients. A list of terms is returned which have to be
        summed over to produce the coupling Hamiltonian.
        In case of RWA, right now only 2 subsystems are supported.

        Returns
        -------
        list[list[Array]]
            A list of operators for each subsystem.
            The subsystems are the outer list.

        """
        if self._is_longitudinal:
            # Number operator (a^\dagger a) for each subsystem
            return [[jnp.diag(jnp.arange(0, s, dtype=jnp.float64)) for s in self._dims]]

        elif self._use_rwa:
            annihilation_op = self._annihilation_op
            return [
                [annihilation_op[0], annihilation_op[1].T],
                [annihilation_op[0].T, annihilation_op[1]],
            ]
        else:
            # (a + a^\dagger) for each subsystem
            return [[(a + a.T) for a in self._annihilation_op]]



"""Coupling Hamiltonian in the rotating frame of drive."""

import jax.numpy as jnp
from jax import vmap

from paraqeet.model.coupling import TwoBodyCoupling
from paraqeet.model.differentiable_hamiltonian import DifferentiableHamiltonian
from paraqeet.quantity import Array, Quantity


class RotatingFrameCoupling(TwoBodyCoupling):
    """Implements the coupling in the rotating frame of the drive.

    If multiple subsystems are coupled specify the difference frequency of the
    individual drive frames.

    NOTE - Right now this only works for TWO SUBSYSTEMS.
    TODO - Generalize this for multiple subsystems

    Parameters
    ----------
    subsystems: list[Hamiltonian]
        A set of Hamiltonians which represent the coupling
    coefficient: Quantity
        Constant drive coefficient.
    diffFreq: Quantity
        Difference of drive frequencies for multiple subsystems.
    """

    _subsystem_A: DifferentiableHamiltonian
    _subsystem_B: DifferentiableHamiltonian
    _coefficient: Quantity
    _diff_freq: Quantity

    def __init__(
        self,
        subsystem_A: DifferentiableHamiltonian,
        subsystem_B: DifferentiableHamiltonian,
        coefficient: Quantity,
        diffFreq: Quantity,
    ):
        super().__init__(subsystem_A, subsystem_B, coefficient, is_longitudinal=False)
        self._diff_freq = diffFreq

    def get_parameters(self) -> list[Quantity]:
        """Return the coupling coeffecient and the difference frequency.

        NOTE - Optimization using relational quantities can be optimize the
        drive frequencies for the two subsystems.

        Parameters
        ----------
        list[Quantity]
            Returns the list of parameters of the system.

        """
        return [self._coefficient, self._diff_freq]

    def _coupling_operators(self) -> list[Array]:
        """Return the annihilation operator. Special implementation for two subsystems."""
        if len(self.subsystems) > 2:
            raise NotImplementedError("No implementation for more than 2 subsystems.")
        dim = self.subsystem_A.dimension()
        annihilation_ops: list[Array] = [jnp.sqrt(jnp.diag(jnp.arange(1, dim), k=1))]
        dim = self.subsystem_B.dimension()
        annihilation_ops.append(jnp.sqrt(jnp.diag(jnp.arange(1, dim), k=1)).conj().T)
        return annihilation_ops

    # TODO: is t a one time point or an array of time points?
    def get_RWA_couplings(self, t: Array) -> list[list[Array]]:
        """Return the matrix representation of the coupling for all subsystems.

        A list of terms in the coupling is returned, where each of the term
        contains operators for each subsystem. A composite Hamiltonian puts
        these operators in the correct position in the tensor space to create
        the operators and then sum over the terms.

        Parameters
        ----------
        t: Array
            One time step.

        Returns
        -------
        list[list[Array]]
            The outer list are the coupling terms. The inner list contains
            matrices for each subsystem. The matrices (Array) have the same
            shape as the subsystem's Hamiltonian.get_value_one_time: (n,n)
            with n the subsystem dimension.
        """
        annihilation_ops = self._coupling_operators()

        annihilation_ops[0] *= self._coefficient.get_value() * jnp.exp(1j * self._diff_freq.get_value() * t)
        annihilation_ops_conj = [a.conj().T for a in annihilation_ops]
        return [annihilation_ops, annihilation_ops_conj]

    def get_RWA_gradients(self, times: Array):
        """Compute the gradients of the coupling expression in the rotating frame, i.e.
        include a phase factor for several timesteps.
        """
        # ignoring mypy due to vmap
        return vmap(self.get_RWA_gradients_one_time)(times)  # type: ignore

    # TODO: should we rename this method to get_RWA_gradient_at_timestep for consistency?
    def get_RWA_gradients_one_time(self, t: float) -> list[list[list[Array]]]:
        """Get the one-time gradient of the matrix.

        Returns the gradient of the matrix representation of the coupling
        for all subsystems. Each entry in the list is the gradient with
        respect to one parameter, factorised into subsystems
        (representing a list of term in the coupling).

        Parameters
        ----------
        t: Array
            One time point.

        Returns
        -------
        list[list[list[Array]]]
            The outer list represents the gradients with respect to
            all optimized parameters. The rest is in the same shape as the
            result of get_matrices_one_time.

        """
        annihilation_ops = self._coupling_operators()
        if self._is_optimized(self._coefficient):
            annihilation_ops[0] *= jnp.exp(1j * self._diff_freq.get_value() * t)
            annihilationOps_conj = [a.conj().T for a in annihilation_ops]
            grads = [[annihilation_ops, annihilationOps_conj]]
        elif self._is_optimized(self._diff_freq):
            annihilation_ops[0] *= self._coefficient.get_value() * 1j * t
            annihilationOps_conj = [a.conj().T for a in annihilation_ops]
            grads = [[annihilation_ops, annihilationOps_conj]]
        else:
            grads = [[[jnp.zeros_like(ann_op) for ann_op in annihilation_ops]] * 2]
        return grads



"""Class definition of the 7th-order Verner ODE solver for GRAPE."""

from functools import partial

import jax
import jax.numpy as jnp
from jax import jit, vmap
from jax.lax import dynamic_slice_in_dim, scan

from paraqeet.exceptions import ConfigurationException
from paraqeet.model.equation_of_motion import EquationOfMotion
from paraqeet.propagation.differentiable_propagation import DifferentiablePropagation
from paraqeet.propagation.vern7 import Vern7
from paraqeet.quantity import Array

jax.config.update("jax_enable_x64", True)


class Vern7GRAPE(Vern7, DifferentiablePropagation):
    """
    Solve EOMs by 7th order ODE method to compute gradients using GRAPE.

    Compute the gradients of a quantum system for PWC pulses by using GRAPE.
    Here, we use forward propagation of the initial state and backward
    propagation of the target state to compute the gradients.

    The state propagations are done by the `Vern7 ODE` method.

    _resolution: float
        Simulation resolution.
    _initial_state: Array = None
        Initial state for forward propagation.
    _target_state: Array = None
        Target state for backward propagation.
    """

    _target_state: Array | None = None

    def __init__(self, model: EquationOfMotion, resolution: float):
        super().__init__(model, resolution)

        if self.is_open:
            self._reverse_step_function = self._reverse_lindblad_step
        else:
            self._reverse_step_function = self._reverse_schrodinger_step

    def set_target_state(self, target_state: Array) -> None:
        """Set target state for backward propagation.

        Parameters
        ----------
        target_state: Array
            Target state.
        """
        # For open system check if target state is a density matrixs.
        if self.is_open:
            if target_state.shape[-1] != target_state.shape[-2]:
                raise ConfigurationException(
                    f"Obtained a state vector of shape {target_state.shape} as target state. "
                    + "For open system propagation expected a density matrix as the target state."
                )

        self._target_state = target_state

    def _reverse_schrodinger_step(self, state: Array, h: Array, cols: list[Array]):
        return jnp.matmul(state, h)

    def _reverse_lindblad_step(self, state: Array, h: Array, cols: list[Array]):
        del_rho = Vern7._commutator(h, state)
        for col in cols:
            del_rho -= jnp.matmul(jnp.matmul(Vern7._dagger(col), state), col)
            del_rho += 0.5 * Vern7._anti_commutator(jnp.matmul(Vern7._dagger(col), col), state)
        return del_rho

    @partial(jit, static_argnums=(0,))
    def _forward_and_backward_propagation(
        self,
        psis_t,
        lamdas_t,
        eom,
        col,
        steps_arr,
    ):
        """Forward propagate initial state and backward propagate target state.

        JIT compiled and uses `jax.lax.scan` to avoid compilation overhead.

        Parameters
        ----------
        psis_t: Array
            Forward propagated state
        lamdas_t: Array
            Backward propagated state
        """

        def forward_propagation(psis_t, index):
            psis_t = self._vern7_one_step(
                psis_t,
                dynamic_slice_in_dim(eom, start_index=9 * index, slice_size=9, axis=0),
                col,
            )
            return psis_t, psis_t

        def backward_propagation(lamdas_t, index):
            lamdas_t = self._vern7_one_step(
                lamdas_t,
                dynamic_slice_in_dim(eom, start_index=9 * index, slice_size=9, axis=0),
                col,
            )
            return lamdas_t, lamdas_t

        psis_t, _ = scan(forward_propagation, psis_t, steps_arr)

        self.step_function = self._reverse_step_function
        eom = (-1) * jnp.flip(eom, axis=0)
        lamdas_t, _ = scan(backward_propagation, lamdas_t, steps_arr)

        return psis_t, lamdas_t

    def get_value_and_gradient(self, times: Array) -> tuple[Array, Array]:
        """Compute gradients using GRAPE.

        Compute the forward propagation of the initial state and
        the backward propagation of the target state.

        Psis represent the forward propagation and lamdas represent
        the backward propagation states.

        This propagation method assumes a PWC pulse as input.

        Note: This method only computes the first order gradients right now.
        """
        if len(times) < 2:
            raise ValueError("Vern7GRAPE.get_value_and_gradient needs at least two time points.")

        if self._initial_state is None:
            raise ConfigurationException("Initial state is not set")

        if self._target_state is None:
            raise ConfigurationException("Target state is not set")

        init_state = jnp.array(self._initial_state, dtype=jnp.complex128)
        target_state = jnp.array(self._target_state, dtype=jnp.complex128)
        target_state = target_state.conj().T

        if self._model is None:
            raise ConfigurationException("No equation of motion is configured.")
        eom_func = self._model.get_value
        grad_func = self._model.get_value_and_gradient

        # Verify if `model.ode_propagation` is set to `True`.
        # ode_propgation returns hamiltonian and collapse operators separately.
        if not self._model.ode_propagation:
            raise ConfigurationException("Please set `model.ode_propagation` to `True` for this propagation method.")

        psis_list = [init_state]
        lamdas_list = [target_state]

        for ti in range(1, len(times)):
            psi_t = psis_list[ti - 1]
            lamda_t = lamdas_list[ti - 1]

            # Interpolate times
            time_grid, dt = self._construct_times(times, ti)
            times_interp = Vern7._interpolate_time(time_grid, dt)
            times_interp = times_interp + dt / 2

            if len(times_interp) < 9:
                raise ConfigurationException(
                    "Propagation resolution has been set very low. Higher resolution needed for this method."
                )

            # TODO: currently separate time grids are required for the EOM and the gradients.
            # TODO: Can we use one so that the value and gradients are computed simultaneously?

            eom, cols = eom_func(times_interp)

            psi_t, lamda_t = self._forward_and_backward_propagation(
                psi_t, lamda_t, eom * dt, jnp.array(cols) * jnp.sqrt(dt), jnp.arange(0, len(time_grid), 1)
            )

            psis_list.append(psi_t)
            lamdas_list.append(lamda_t)

        psis = jnp.array(psis_list)
        lamdas = jnp.array(lamdas_list)

        lamdas = jnp.flip(lamdas, axis=0)
        _, dh_dps = grad_func(times[:-1] + dt / 2)
        dh_dps = jnp.array(dh_dps) * dt

        grads = []
        n_params = dh_dps.shape[1]
        for i in range(n_params):
            if self.is_open:
                fwd_prop_state = vmap(Vern7._commutator, in_axes=(0, 0))(dh_dps[:, i, ...], psis[1:])
            else:
                fwd_prop_state = vmap(jnp.matmul, in_axes=(0, 0))(dh_dps[:, i, ...], psis[1:])

            grad = vmap(jnp.matmul, in_axes=(0, 0))(lamdas[1:], fwd_prop_state)
            grad = jnp.squeeze(grad)
            grads.append(grad)
        return psis, jnp.array(grads)



"""Test the mixed state transfer fidelity."""

import numpy as np
import pytest

from paraqeet.measurement.mixed_state_transfer_fidelity import (
    MixedStateTransferFidelity,
)
from tests.propagation.identity_propagation import IdentityPropagation
from tests.propagation.random_propagation import RandomPropagation


def random_mixed_state(dimension):
    """Generate random mixed states."""
    state = np.random.random((dimension, dimension)) + 1j * np.random.random((dimension, dimension))
    state = state @ np.conjugate(state.T)
    return state / np.trace(state)


def test_limits_vectors():
    """Test state vector limits.

    The fidelity for state vectors is always in the interval [0, 1).

    """
    for size in range(2, 30):
        targetState = random_mixed_state(size)
        propagation = RandomPropagation(size, True)
        times = np.array([1.0])
        measurement = MixedStateTransferFidelity(propagation, targetState)

        for i in range(100):
            m = measurement.measure(times)
            assert 0.0 <= m <= 1.0


def test_vector_equality():
    """Test that F(v,v) = 1 for state vectors."""
    for size in range(2, 30):
        for i in range(100):
            state = random_mixed_state(size)
            propagation = IdentityPropagation()
            propagation.set_initial_state(state)
            measurement = MixedStateTransferFidelity(propagation, state)
            m = measurement.measure(times=np.array([1.0]))
            np.testing.assert_almost_equal(m, 1.0, decimal=2)


def test_incompatible_shape():
    """Test incompatible shape of propagators.

    The propagators which have a different dimension than the
    target state should raise an exception.

    Raises
    ------
    Exception
        Different dimensions of the propagators raise an exception.

    """
    allDims = np.arange(2, 30)
    for dim in allDims:
        for i in range(100):
            targetState = random_mixed_state(dim)

            # create a propagator of a different dimension
            dimensions = np.delete(allDims, np.where(allDims == dim)[0][0])
            propagation = RandomPropagation(np.random.choice(dimensions), True)

            measurement = MixedStateTransferFidelity(propagation, targetState)
            with pytest.raises(Exception):
                measurement.measure(times=np.array([1.0]))



"""Test the random propagation model."""

from functools import partial

import jax.numpy as jnp
import numpy as np
from jax import jit
from scipy.stats import unitary_group

from paraqeet.differentiable import Differentiable
from paraqeet.propagation.propagation import Propagation
from paraqeet.quantity import Array, Quantity
from tests.model.dummy_model import DummyEquationsOfMotion
from tests.model.empty_hamiltonian import EmptyHamiltonian


class RandomPropagation(Propagation, Differentiable):
    """Mock random propagation implementation.

    Returns random state vectors, density matrices, or propagators.

    Parameters
    ----------
    dimension: int
        Hilbert space size for the generated states.
    generate_matrices : bool, default=False
        Whether to generate matrices instead of vectors.
    auto_update: bool, default=True
        Whether to return a new random state at every call of propagate.
        If false, propagate will return the same state until update was called.
    """

    _dimension: int
    _create_matrices: bool
    _auto_update: bool
    _state: Array

    def __init__(
        self,
        dimension: int,
        generate_matrices: bool = False,
        auto_update: bool = True,
    ):
        super().__init__(DummyEquationsOfMotion(EmptyHamiltonian(0)), 1e9)
        self._dimension = dimension
        self._create_matrices = generate_matrices
        self._auto_update = auto_update
        self.update()
        self.is_open = False

    def get_parameters(self) -> list[Quantity]:
        """Returns an empty list."""
        return []

    def set_initial_state(self, state: Array):
        """Set the initial state of the system.

        Set it to the given state.

        Parameters
        ----------
        state: Array
            Given state to set as the initial state.

        """
        pass

    def propagate(self, time: Array) -> Array:
        """Propagate the system through time.

        Parameters
        ----------
        time: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the updated state of the system.

        """
        if self._auto_update:
            self.update()
        return jnp.array([self._state] * len(time))

    def get_value_and_gradient(self, times: Array) -> tuple[Array, Array]:
        # Returns an empty gradient because the class has 0 parameters
        empty_gradient = jnp.zeros(shape=(len(times), 0, len(self._state)))
        return self.propagate(times), empty_gradient

    @staticmethod
    @partial(jit, static_argnums=(0,))
    def __create_random_dm(dim: int, rho: Array):
        rho /= jnp.trace(rho)
        u = unitary_group.rvs(dim)
        return jnp.conjugate(u.T) @ rho @ u

    @staticmethod
    @jit
    def __create_random_vec(state: Array):
        return state / jnp.sqrt(jnp.vdot(state, state))

    def update(self) -> None:
        """Update the state on propagation.

        Makes sure that the next call to propagate will return a
        new random state.

        """
        if self._create_matrices:
            # generate a random density matrix by rotating a
            # random diagonal matrix
            rho = jnp.diag(np.random.random(self._dimension))
            self._state = RandomPropagation.__create_random_dm(self._dimension, rho)
        else:
            # generate a random state vector
            state = np.random.random((self._dimension, 1)) + 1j * np.random.random((self._dimension, 1))
            self._state = RandomPropagation.__create_random_vec(state)



"""Class definition of a closed model."""

from collections.abc import Callable

import jax.numpy as jnp

from paraqeet.model.differentiable_hamiltonian import DifferentiableHamiltonian
from paraqeet.model.equation_of_motion import EquationOfMotion
from paraqeet.quantity import Array, Quantity


class ClosedSystem(EquationOfMotion):
    """Model of a closed physical system, defined by a Hamiltonian.

    Its dynamics is given by the Schrödinger equation.

    Parameters
    ----------
    hamiltonian : Hamiltonian
        Matrix representation of a Hamiltonian.

    """

    _get_value_method: Callable

    def __init__(self, hamiltonian: DifferentiableHamiltonian, ode_propagation: bool = False):
        super().__init__(hamiltonian)
        self.ode_propagation = ode_propagation

    @property
    def ode_propagation(self) -> bool:
        """Flag to set method of propagation to ODE.

        Returns
        -------
        bool
            Flag to use ODE propagation.
        """
        return self._ode_propagation

    @ode_propagation.setter
    def ode_propagation(self, ode_propagation: bool) -> None:
        self._ode_propagation = ode_propagation

        if ode_propagation:
            self._get_value_method = self._get_ode_propagation_eom
        else:
            self._get_value_method = self._get_eom

    def get_parameters(self) -> list[Quantity]:
        """Get a list of optimizable parameters.

        Returns
        -------
        List[Quantity]
            List of optimizable parameters of the system.

        """
        return self._hamiltonian.get_parameters()

    def _get_eom(self, times: Array) -> Array:
        """Get the matrix equations of motion.

        Computes the right hand side of the Schrödinger equation
        without multiplying the state.
        Used for unitary solvers.

        Parameters
        ----------
        times : Array
            Vector of time samples.

        Returns
        -------
        Array
            RHS with dimension [t, n, n]  with 't' as time
            and 'n' as Hilbert space dimension.

        """
        return -1.0j * self._hamiltonian.get_value(times)

    def _get_ode_propagation_eom(self, times: Array) -> tuple[Array, Array]:
        """Get the matrix equations of motion for ODE solver.

        Here we return an empty array for the collapse operator.
        """
        return -1.0j * self._hamiltonian.get_value(times), jnp.empty((1,), dtype=jnp.complex128)

    def get_value(self, times: Array):
        """Get the matrix equations of motion.

        Computes the right hand side of the Schrödinger equation
        without multiplying the state.
        Used for unitary solvers.

        Parameters
        ----------
        times : Array
            Vector of time samples.

        Returns
        -------
        Array
            RHS with dimension [t, n, n]  with 't' as time
            and 'n' as Hilbert space dimension.

        """
        return self._get_value_method(times)

    def get_value_and_gradient(self, times) -> tuple[Array, Array]:
        """Compute the gradient of getMatrix.

        Parameters
        ----------
        times : Array
            Vector of time samples.

        Returns
        -------
        Array
            Returns the gradient of getMatrix.

        """
        eom, eom_gradient = self._hamiltonian.get_value_and_gradient(times)
        return -1.0j * eom, -1.0j * eom_gradient



"""Class definition of the pulse smoothness. It follows the definition in
[Heeres2017], in particular Eqs. 21 of the supplementary material.

References
[Heeres2017] R. Heeres et al., Nat. Comm. 8, 94 (2017)
"""

import jax
import jax.numpy as jnp

from paraqeet.differentiable import Differentiable
from paraqeet.measurement.measurement import NormalizableMeasurement
from paraqeet.quantity import Array
from paraqeet.signal.pwc_generator import PWCGenerator

jax.config.update("jax_enable_x64", True)


class Smoothness(NormalizableMeasurement, Differentiable):
    """Smoothness of a pulse. It follows the definition in
    Heeres et al., https://arxiv.org/abs/1608.02430 (2017), in particular
    Eqs. 23 and 24 of the supplementary material.

    Parameters
    ----------
    pwc_generator: PWCGenerator
        The generator from which we extract the pulse.
    times: Array
        One-dimensional vector of timestamps.
    """

    _pwc_generator: PWCGenerator

    def __init__(self, pwc_generator: PWCGenerator):
        # super().__init__(pwc_generator.tlist)
        self._pwc_generator = pwc_generator

    def measure(self, times: Array) -> Array | float:
        """Return measurement in the range [0, 1]."""
        return self.calculate_normalized_scalar(times)

    # TODO: This should depend on the internal time grid and not on the input time value.
    # This means that the `times` should just be a float.
    def calculate_normalized_scalar(self, times: Array | float) -> float:
        """Returns the normalized sum of consecutive square differences of the pulse.
        As the maximums difference is twice the maximum amplitude, the normalization
        factor is the number of piecewise constants minus 1 time sthe maximum
        difference squared.

        Returns
        -------
        float
            The normalized sum of consecutive square differences in the pulse.
        """
        pulse = self._pwc_generator.get_value(times)
        num_pwc = jnp.shape(pulse)[0]

        norm_coeff = (num_pwc - 1) * (2 * self._pwc_generator.max_amplitude) ** 2

        def get_squared_difference(index):
            """Squared difference between two consecutive bins in PWC pulse.

            Parameters
            ----------
            index : int
                index of the bin
            """
            return jnp.abs(pulse[index] - pulse[index + 1]) ** 2

        indices = jnp.arange(0, num_pwc - 1)
        vmap_get_squared_difference = jax.vmap(get_squared_difference)
        return float(1.0 - jnp.sum(vmap_get_squared_difference(indices)) / norm_coeff)

    def get_value_and_gradient(self, times: Array) -> tuple[Array, Array] | tuple[float, Array]:
        """Measure with gradient.

        Compute the measurement value as in measure_normalized_scalar()
        and the gradient with respect to all parameters in the optimization map.
        For parameters that are not in the passed PWCGenerator the partial derivative
        if simply zero.

        Returns
        -------
        Tuple[float, Array]
            Tuple of function value as bare float and gradient of shape (n_parameters,)

        """
        opt_pwc_params = self._pwc_generator.optimizable_parameters
        opt_params = self._pwc_generator.all_optimizable_parameters

        def get_partial_derivative(n, vec):
            """Derivatives of the smoothness measure for 3 cases: starting point, center and end point.

            Parameters
            ----------
            n : int
                location in the piecewise constant vector
            vec : Array
                piecewise constant vector
            """
            num_pwc = vec.shape[0]
            res = (
                ((2 * vec[n]) - vec[n + 1] - vec[n - 1])
                * (1.0 - jnp.where(n == 0, 1.0, 0.0))
                * (1.0 - jnp.where(n == num_pwc - 1, 1.0, 0.0))
            )
            res += (vec[0] - vec[1]) * jnp.where(n == 0, 1.0, 0.0)
            res += (vec[num_pwc - 1] - vec[num_pwc - 2]) * jnp.where(n == num_pwc - 1, 1.0, 0.0)
            return -res

        if len(opt_params) > 0:
            grad_list = []
            for param in opt_params:
                num_pwc = param.get_value().shape[0]
                norm_coeff = (num_pwc - 1) * (2 * self._pwc_generator.max_amplitude) ** 2
                if id(param) in [id(pwc_param) for pwc_param in opt_pwc_params]:
                    n_vec = jnp.arange(num_pwc)
                    vmap_get_partial_derivative = jax.vmap(get_partial_derivative, in_axes=(0, None))
                    grad_list.append(2 * vmap_get_partial_derivative(n_vec, param.get_value()) / norm_coeff)
                else:
                    grad_list.append(jnp.zeros(num_pwc))
            gradient = jnp.concatenate(grad_list)
        else:
            gradient = jnp.empty((1, 0))  # this is purely conventional

        return self.calculate_normalized_scalar(times), gradient



"""Class definition of the empty Hamiltonian for testing."""

import numpy as np

from paraqeet.model.hamiltonian import Hamiltonian
from paraqeet.quantity import Array, Quantity


class EmptyHamiltonian(Hamiltonian):
    """A Hamiltonian that is filled with zeros for all time steps.

    Parameters
    ----------
    dimension : int
        The dimension for the representation of the Hamiltonian.
    """

    _dimension: int

    def __init__(self, dimension: int):
        super().__init__([])

        self._dimension = dimension

    def get_value(self, times: Array) -> Array:
        """Get the matrix representation of the Hamiltonian.

        Parameters
        ----------
        t: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            The matrix representation of the Hamiltonian.

        """
        return np.zeros((len(times), self._dimension, self._dimension))

    def get_parameters(self) -> list[Quantity]:
        """ """
        return []

    # TODO: implement abstract methods from Hamiltonian
    def dimension(self) -> int:
        raise NotImplementedError("Method not implemented yet.")

    # TODO: implement abstract methods from Hamiltonian
    def get_value_at_timestep(self, timestep: Array) -> Array:
        raise NotImplementedError("Method not implemented yet.")

    # TODO: implement abstract methods from Hamiltonian
    def get_gradient_at_timestep(self, t: Array) -> Array:
        raise NotImplementedError("Method not implemented yet.")

    # TODO: implement abstract methods from Hamiltonian
    def get_collapseops(self) -> list[tuple[Array, Array]]:
        raise NotImplementedError("Method not implemented yet.")



"""Class definition for a mock system that always returns the same value."""

import jax.numpy as jnp

from paraqeet.differentiable import Differentiable
from paraqeet.measurement.measurement import Measurement
from paraqeet.propagation.propagation import Propagation
from paraqeet.quantity import Array


class ConstantMeasurement(Measurement, Differentiable):
    """Mock implementation that always returns the same value.

    Parameters
    ----------
    propagation : Propagation
        Abstract base class for any implementation that can solve
        the equation of motion.
    value : float, default=1.0
        Value of the measurement.
    times : float | None, optional
        Time variable value.
    """

    _propagation: Propagation
    _value: Array

    def __init__(
        self,
        propagation: Propagation,
        value: Array = jnp.array(1.0),
    ):
        self._propagation = propagation
        self._value = value

    def measure(self, times: Array) -> Array:
        """Get the measurement value.

        Returns
        -------
        Array
            The value of the measurement.

        """
        return self._value

    def get_value_and_gradient(self, times: Array) -> tuple[Array, Array] | tuple[float, Array]:
        """Get measurement value and gradient"""
        grad = jnp.array([self._value, 0.0])
        return self._value, grad



"""Class definition of the Scipy piecewise exponential propagation model.

Uses the GOAT optimization method.

"""

from functools import partial

import jax.numpy as jnp
from jax import jit, vmap
from jax.lax import scan

from paraqeet.exceptions import ConfigurationException
from paraqeet.propagation.differentiable_propagation import DifferentiablePropagation
from paraqeet.propagation.scipy_expm import ScipyExpm
from paraqeet.quantity import Array


class ScipyExpmGOAT(ScipyExpm, DifferentiablePropagation):
    """Solve EOMs by piecewise exponentiation via Scipy using GOAT."""

    def _create_super_state(self, psi: Array, dpsis: Array) -> Array:
        """Create a state for the system state and also for gradient vectors.

        Parameters
        ----------
        psi : Array
            State of the system.
        dpsis : Array
            Differential of state.

        Returns
        -------
        Array
            Returns a super state created from the state and the differential.

        """
        super_state = [psi]
        super_state.extend(dpsis)
        psi_t = jnp.concatenate(super_state)
        return psi_t

    def _create_goat_ham(self, n_params, eom, grads):
        """Create a Hamiltonian for the GOAT optimization method.

        Parameters
        ----------
        n_params: int
            Number of parameters.
        eom: Array
            Equations of motion in matrix form.
        grads: Array
            Gradients of the system at a particular step.

        Returns
        -------
        Array
            Hamiltonian for the GOAT optimization method.

        """
        line = [eom]
        zeros_like_eom = jnp.zeros_like(eom)
        line.extend([zeros_like_eom] * n_params)
        goat_ham_list = [line]
        for ii, dh_dp in enumerate(grads, start=1):
            line = [dh_dp]
            line.extend([zeros_like_eom] * (ii - 1))
            line.append(eom)
            line.extend([zeros_like_eom] * (n_params - ii))
            goat_ham_list.append(line)

        return jnp.block(goat_ham_list)

    @partial(jit, static_argnums=(0, 1))
    def _propagate_gradient(self, n_params, psis_t, eom, grads, steps_arr):
        def propagate_body(psis_t, index):
            goat_ham = self._create_goat_ham(n_params, eom[index], grads[index])
            psis_t = ScipyExpm._propagate_psi(goat_ham, psis_t)
            return psis_t, psis_t

        psis_t, _ = scan(propagate_body, psis_t, steps_arr)
        return psis_t

    def get_value_and_gradient(self, times: Array) -> tuple[Array, Array]:
        """Solve the GOAT equation for the gradient vector.

        Parameters
        ----------
        time: Array
            Array of timesteps.

        Returns
        -------
        tuple[Array, Array]
            First dimension is time, second dimension is the parameter.

        """
        if len(times) < 2:
            raise ValueError("ScipyExpmGOAT.get_value_and_gradient needs at least two time points.")

        if self._initial_state is None:
            raise ConfigurationException("Initial state is not set")
        if self._model is None:
            raise ConfigurationException("No equation of motion is configured.")
        n_params = self._model.get_value_and_gradient(jnp.array([0.0]))[1].shape[1]
        dim = self._initial_state.shape[0]
        psis = [jnp.array(self._initial_state, dtype=jnp.complex128)]
        dpsis: list[Array] = [jnp.zeros((n_params,) + self._initial_state.shape, dtype=jnp.complex128)]

        grad_func = self._model.get_value_and_gradient

        for ti in range(1, len(times)):
            times, dt = self._construct_times(times, ti)
            psi_t = self._create_super_state(psis[-1], dpsis[-1])

            eom, grads = grad_func(times + dt / 2)
            eom = eom * dt
            grads = jnp.array(grads) * dt

            psi_t = self._propagate_gradient(n_params, psi_t, eom, grads, jnp.arange(0, len(times), 1))
            psis.append(jnp.array(psi_t[0:dim]))
            dpsis.append(jnp.array([psi_t[dim * ii : dim * (ii + 1)] for ii in range(1, n_params + 1)]))

        psis_arr = jnp.array(psis)
        dpsis_arr = jnp.array(dpsis)

        if self.is_open:
            dim = int(jnp.sqrt(eom.shape[-1]))
            psis_arr = vmap(ScipyExpm._convert_vec_to_dm, in_axes=(0, None))(psis_arr, dim)
            dpsis_arr = vmap(vmap(ScipyExpm._convert_vec_to_dm, in_axes=(0, None)), in_axes=(0, None))(dpsis_arr, dim)
        return psis_arr, dpsis_arr



from functools import partial

import jax.numpy as jnp
from jax import jit, vmap
from jax.scipy.special import erf

from paraqeet.quantity import Array, Quantity
from paraqeet.signal.generator import Generator
from paraqeet.signal.waveform import Waveform


class PWCGenerator(Generator):
    """Convert a complex envelope to PWC pulse.

    This sets the pulse parameters to the `tlist` points.
    The gradient of the pulse wrt the PWC bins is 1 at that time point and zero
    everywhere else.

    This Generator doesn't add the LO signal to the envelope pulse.
    Driving with a PWC pulse (without the LO) is usually done in the rotating
    frame of drive.

    This Generator converts the input complex pulse to the 'in-phase' and
    'out-of-phase' components. This naming convention is used by following [Krantz2019].
    In the literature of signal processing these are also called 'in-phase' and 'quadrature'
    components (refer to https://en.wikipedia.org/wiki/In-phase_and_quadrature_components).

    [Krantz2019] Krantz et al., “A Quantum Engineer’s Guide to Superconducting Qubits.” Applied Physics Reviews 6(2019).

    _envs: list[Waveform]
        List of Envelopes
    _tlist: Array
        Left time points for discretization. These can be used for propagation and optimization.
    _time_grid: Array
        Time grid used to discretize the pulse. These are shifted from tlist by dt, and doesn't include zero time.
    _max_amplitude: float
        Maximum amplitude of the drive
    _inphase: Quantity
        The in-phase component of the pulse
    _outofphase: Quantity
        The out-of-phase component of the pulse
    _optimizable_paramters: list[Quantity]
        List of own parameters that would be optimized by the optimizer.
    _multiply_flat_top: bool
        Flag to multiply flat-top-Gaussain pulse to the signal to ensure it
        starts and ends at zero.

    Parameters
    ----------
    envelopes : List[Waveform]
        List of input devices.

    """

    _envs: list[Waveform]
    _tlist: Array
    _time_grid: Array
    _max_amplitude: float
    _inphase: Quantity
    _outofphase: Quantity
    _optimizable_parameters: list[Quantity] = []
    _multiply_flat_top: bool = False

    def __init__(
        self,
        envelopes: list[Waveform] | None,
        tlist: Array,
        max_amplitude: float | None = None,
    ):
        self._envs = envelopes or []
        self._tlist = tlist

        # Choose the center point for time grid
        dt = tlist[1] - tlist[0]
        self._time_grid = tlist[:-1] + dt / 2

        if max_amplitude is not None and max_amplitude < 0.0:
            raise ValueError("The maximum drive amplitude must be positive.")
        elif max_amplitude is None:
            env = self._compute_shape()
            self._max_amplitude = 2 * float(jnp.max(jnp.abs(env)))
        else:
            self._max_amplitude = max_amplitude

        self._setup_inphase_and_outofphase()
        self._t_final = self._time_grid[-1]

    @partial(jit, static_argnums=(0,))
    def _compute_envelope(self, t):
        t_final = self._t_final
        ramp_time = t_final / 25
        ramp_up = 1 + erf((t - 2 * t_final / 20) / ramp_time)
        ramp_down = 1 + erf((-t + 18 * t_final / 20) / ramp_time)
        return ramp_up * ramp_down / 4

    @property
    def tlist(self) -> Array:
        """Get time grid discretization for generating PWC pulse.

        Returns
        -------
        Array
            Array of time points at which envelope is discretized.
        """
        return self._tlist

    @tlist.setter
    def tlist(self, tlist: Array) -> None:
        """Set time grid discretization for generating PWC pulse.

        Parameters
        ----------
        tlist : Array
            Array of time points at which envelope is discretized.
        """
        self._tlist = tlist

        # update time grid
        dt = tlist[1] - tlist[0]
        self._time_grid = tlist[:-1] + dt / 2

        self._setup_inphase_and_outofphase()

    @property
    def max_amplitude(self) -> float:
        """Get the maximum drive amplitude.

        Returns
        -------
            The value of the maximum drive amplitude.
        """
        return self._max_amplitude

    @max_amplitude.setter
    def max_amplitude(self, max_amplitude: float) -> None:
        """Set the maximum drive amplitude of the drive.

        Parameters
        ----------
        max_amplitude: float
            The value of the maximu drive amplitude.
        """
        self._max_amplitude = max_amplitude
        self._setup_inphase_and_outofphase()

    @property
    def envs(self) -> list[Waveform]:
        """Gets the list of envelopes.

        Returns
        -------
        list[Waveform]
            The list of waveforms associated with the generator.
        """
        return self._envs

    @property
    def multiply_flat_top(self) -> bool:
        """Flag to multiply the pulse with a FlatTop.

        This can be used to make the start and end values zeros and force the
        PWC pulse to change smoothly.

        Returns
        -------
        multiply_flat_top: bool
            Flag value for multiply_flat_top.
        """
        return self._multiply_flat_top

    @multiply_flat_top.setter
    def multiply_flat_top(self, multiply_flat_top: bool) -> None:
        """Set flag to multiply the pulse with a FlatTop.

        This can be used to make the start and end values zeros and force the
        PWC pulse to change smoothly.

        Parameters
        ----------
        multiply_flat_top : bool
            Flag value for multiply_flat_top.
        """
        self._multiply_flat_top = multiply_flat_top
        self._setup_inphase_and_outofphase()

    def get_number_of_pwc_pixels(self) -> int:
        """Return number of PWC pixels generated by the PWC generator."""
        return len(self.tlist)

    def _compute_shape(self) -> Array:
        env = jnp.zeros_like(self._time_grid, dtype=jnp.complex128)
        for dev in self._envs:
            env += dev.get_value(self._time_grid)
        return env

    def _setup_inphase_and_outofphase(self) -> None:
        """Generate inphase and outphase Quantities using tlist."""
        env = self._compute_shape()

        # max_abs = jnp.max(jnp.abs(env))
        max_component = self._max_amplitude / jnp.sqrt(2)
        bound = max_component * jnp.ones_like(self._time_grid)

        self._inphase = Quantity(
            jnp.real(env),
            min_value=-bound,
            max_value=bound,
            name="Inphase",
        )
        self._outofphase = Quantity(
            jnp.imag(env),
            min_value=-bound,
            max_value=bound,
            name="out-of-phase",
        )

    def _get_partial_derivatives(self, times) -> Array:
        env_grads = []
        for dev in self._envs:
            _, grad = dev.get_value_and_gradient(times)
            dev_grad = jnp.concat([jnp.real(grad), jnp.imag(grad)])
            env_grads.append(dev_grad)
        return jnp.hstack(env_grads)

    def _update_inphase_and_outofphase(self) -> None:
        env = self._compute_shape()
        self._inphase.set_value(jnp.real(env))
        self._outofphase.set_value(jnp.imag(env))

    def get_parameters(self) -> list[Quantity]:
        """Return a list of parameters.

        Return the inphase and out-of-phase as parameters.

        Returns
        -------
        list[Quantity]
            All Parameters describing the signal.
        """
        return [self._inphase, self._outofphase]

    def set_optimizable_parameters(self, params: list[Quantity]) -> None:
        """Set specified parameters to be optimized.

        Optimizable parameters can be inphase and out-of-phase.

        Parameters
        ----------
        params : list[Quantity]
        """
        super().set_optimizable_parameters(params)

    @partial(jit, static_argnums=(0,))
    def _pwc_signal(
        self,
        inphase: Array,
        outofphase: Array,
        dt: Array,
        t: Array,
    ) -> Array:
        """Generate a signal for a single time point 't'.

        The PWC signal is generated by finiding the closest time point
        and returning the corresponding amplitude value.

        Parameters
        ----------
        inphase: Array
            1-D vector of step values of real part of the PWC signal.
        out-of-phase: Array
            1-D vector of step values of complex part of the PWC signal.
        tlist: Array
            Time bins of the PWC pulse.
        t: Array
            One time point.

        Returns
        -------
        Array
            Returns the PWC signal value at t.

        """
        index = jnp.array(t / dt, int)
        return inphase[index] + 1.0j * outofphase[index]

    def get_value(self, times: Array | float) -> Array:
        """Generate the PWC signal for time(s) 't'.

        Parameters
        ----------
        times: Array
            One-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the signal vector.

        """
        time_grid = self._time_grid
        dt = time_grid[1] - time_grid[0]
        t_arr = jnp.array(times, ndmin=1)
        inphase = self._inphase.get_value()
        outofphase = self._outofphase.get_value()
        if self._multiply_flat_top:
            env = self._compute_envelope(time_grid)
            inphase *= env
            outofphase *= env
        shape = jnp.squeeze(vmap(self._pwc_signal, in_axes=(None, None, None, 0))(inphase, outofphase, dt, t_arr))
        return shape

    def get_value_and_gradient(self, times: Array) -> tuple[Array, Array]:
        """Return signal gradient wrt inphase and out-of-phase.

        This returns a list of ones as the gradient of the envelope wrt a step
        is 1 for that time bin and 0 everywhere else.

        Parameters
        ----------
        times : Array
            Array of time steps.

        Returns
        -------
        Array
            PWC signal gradients.
        """
        t_arr = jnp.array(times, ndmin=1)

        grads = []
        time_grid = self._time_grid
        dt = time_grid[1] - time_grid[0]

        inphase = self._inphase.get_value()
        outofphase = self._outofphase.get_value()

        if self._multiply_flat_top:
            smoothing = self._compute_envelope(time_grid)
            index = jnp.argmin(jnp.abs(jnp.expand_dims(time_grid, axis=1) - t_arr), axis=0)
            env = smoothing[index]
            inphase *= env
            outofphase *= env
        else:
            env = jnp.ones_like(t_arr)

        if self._is_optimized(self._inphase):
            grads.append(env)
        if self._is_optimized(self._outofphase):
            grads.append(1j * env)

        if len(grads) > 0:
            grads_stack = jnp.stack(grads)
        else:
            grads_stack = jnp.empty((t_arr.shape[0], 0))

        shape = jnp.squeeze(vmap(self._pwc_signal, in_axes=(None, None, None, 0))(inphase, outofphase, dt, t_arr))
        return shape, grads_stack

    def get_gradient_at_timestep(self, time: float) -> Array:
        """Return signal gradient wrt inphase and out-of-phase.

        This returns a list of ones as the gradient of the envelope wrt a step
        is 1 for that time bin and 0 everywhere else.

        Parameters
        ----------
        time: Array
            One time step.

        Returns
        -------
        Array
            PWC signal gradients.
        """
        grads = []
        time_grid = self._time_grid

        if self._multiply_flat_top:
            smoothing = self._compute_envelope(time_grid)
            index = jnp.argmin(jnp.abs(time_grid - time))
            env = smoothing[index]
        else:
            env = 1

        if self._is_optimized(self._inphase):
            grads.append(env)
        if self._is_optimized(self._outofphase):
            grads.append(1j * env)

        return jnp.stack(grads, axis=0) if len(grads) > 0 else jnp.empty((0,))



"""Class definition of the Scipy piecewise exponential propagation model."""

from functools import partial

import jax
import jax.numpy as jnp
from jax import jit, vmap
from jax.lax import scan
from jax.scipy.linalg import expm

from paraqeet.exceptions import ConfigurationException
from paraqeet.model.equation_of_motion import EquationOfMotion
from paraqeet.propagation.propagation import Propagation
from paraqeet.quantity import Array

jax.config.update("jax_enable_x64", True)


class ScipyExpm(Propagation):
    """Piecewise matrix exponential propagation system.

    Solve the equation of motion by piecewise exponentiation with the
    Scipy package.

    Parameters
    ----------
    model: Model
        Represents the equation of motion for a given Hamiltonian.
    res: float
        Resolution at which to sample the EOM.

    """

    _resolution: float
    _initial_state: Array | None = None

    def __init__(self, model: EquationOfMotion, resolution: float):
        super().__init__(model, resolution)
        self.resolution = resolution

    def set_initial_state(self, state):
        """Set initial state."""
        # Verify if `model.ode_propagation` is set to `False`.
        # ode_propgation returns hamiltonian and collapse operators separately.
        eom = self._model.get_value(jnp.array([0.0]))
        dim_generator = eom.shape[1]

        if len(eom) == 2:
            raise ConfigurationException("Please set `model.ode_propagation` to `False` for this propagation method.")

        state = self._check_and_fix_state_shape(state, dim_generator)
        self._initial_state = jnp.array(state, dtype=jnp.complex128)

    def _check_and_fix_state_shape(self, state, dim_generator):
        # For closed system check if the initial state has the right dimensions.
        if not self.is_open:
            if len(state.shape) == 1:  # (n,) array
                state = jnp.reshape(state, (-1, 1))

            if state.shape[-2] != dim_generator:
                raise ConfigurationException(
                    f"Obtained a state vector of shape {state.shape} as initial state. "
                    + f"Expected a shape of dimensions `{(dim_generator, 1)}`"
                    + "or an array of initial states of the above dimension."
                )

        # For open system convert Density Matrix to Vectorized form.
        else:
            try:
                if len(state.shape) == 1:  # An (n,) array
                    state = jnp.reshape(state, (-1, 1))

                # Check whether it is a density matrix or vectorized density matrix (or an array of those).
                # Checking if it is a density matrix. Comparing dim (-2) as (0) can be batch dimension
                if state.shape[-2] == jnp.sqrt(dim_generator):
                    # check if it is a square matrix. Check the last 2 dimensions are equal.
                    if state.shape[-1] == state.shape[-2]:
                        # This is a density matrix. Convert to vectorized form.
                        state = ScipyExpm._convert_dm_to_vec(state, int(jnp.sqrt(dim_generator)))

                    # check if it is a list of vectorized density matrices
                    elif state.shape[-1] == dim_generator:
                        # This is an edge case with batch dimension = n, where n is Hilbert space dimension
                        state = jnp.expand_dims(state, -1)

                    else:
                        raise ConfigurationException(
                            "Initial state neither a density matrix nor a vectorized density matrix.\n"
                            + "For a list of vectorized density matrices expected shape is (m, n^2, 1)"
                            + " where m is the batch dimension, n is the Hilbert space dimension."
                        )

                # Not a density matrix (or list). Check if it is a vectorized density matrix.
                elif state.shape[-2] == dim_generator:
                    # vectorized density matrix or a list of vectorized density matrix
                    if state.shape[-1] == state.shape[-2]:
                        # list of vectorized density matrices
                        # This is an edge case: batch dimension = n^2, where n is Hilbert space dimension
                        state = jnp.expand_dims(state, -1)

                    elif state.shape[-1] == 1:
                        # normal vectorized density matrix
                        state = state

                    else:
                        raise ConfigurationException(
                            "Initial state neither a single or list of vectorized density matrix.\n"
                            + "For a list of vectorized density matrices expected shape is (m, n^2, 1)"
                            + " where m is the batch dimension, n is the Hilbert space dimension."
                        )

            except Exception as e:
                raise ConfigurationException(
                    f"Obtained a state vector of shape {state.shape} as initial state. "
                    + "For open system propagation expected a density matrix or vectorized density matrix "
                    + "as the initial state.\n"
                    + f"Raised exception: `{e}`"
                )

        return state

    @staticmethod
    def _convert_dm_to_vec(state_dm: Array, dim: int) -> jnp.ndarray:
        """Helper function to convert a density matrix to vectorized form."""
        vec = jnp.reshape(jnp.transpose(state_dm), (-1, dim**2, 1))
        if vec.shape[0] == 1:
            vec = jnp.squeeze(vec, axis=0)
        return vec

    @staticmethod
    def _convert_vec_to_dm(state_vec: Array, dim: int) -> jnp.ndarray:
        """Helper function to convert a Vectorized density matrix to matrix form."""
        dm = jnp.reshape(state_vec, (-1, dim, dim))
        if dm.shape[0] == 1:
            dm = jnp.squeeze(dm, axis=0)
        return jnp.transpose(dm)

    @partial(jit, static_argnums=(0,))
    def _propagate_in_time(self, psis_t, eom, steps_arr):
        """Propagate the system in time.

        Iteratively propagate state/states (psis_t) according
        to the equation of motion (eom). The eom is exponentiated using
        `jax.scipy.linalg.expm` to compute the propagators.
        The iterations use `jax.lax.scan` to avoid compilation overhead.

        Parameters
        ----------
        psis_t: Array
            State/states at time 't'.
        eom: Array
            Equation of motion for a list of times.
        steps_arr: Array
            Array from 0 to the length of the List of time, in steps of 1
            representing the iteration index.

        Returns
        -------
        Array
            Returns the evolved state.

        """

        def propagate_body(psis_t, index):
            psis_t = ScipyExpm._propagate_psi(eom[index], psis_t)
            return psis_t, psis_t

        psis_t, _ = scan(propagate_body, psis_t, steps_arr)

        return psis_t

    @staticmethod
    @jit
    def _propagate_psi(eom_matrix, psis_t):
        """Propagate the state/states (psis_t).

        Parameters
        ----------
        eom_matrix : Array
            The equations of motion matrix.
        psis_t : Array
            State/states at time 't'.

        Returns
        -------
        Array
            Returns the evolved state.

        """
        return expm(eom_matrix) @ psis_t

    def propagate(self, time: Array) -> Array:
        """Return the solution of the equations of motion.

        Loop over all desired times in time at set resolution.

        Parameters
        ----------
        time: Array
            Any one-dimensional vector of timestamps.

        Returns
        -------
        Array
            Returns the solution of the equations of motion.

        Raises
        ------
        ConfigurationException
            If the initial state is not set.

        """
        if len(time) < 2:
            raise ValueError("ScipyExpm.propagate needs at least two time points.")

        if self._initial_state is None:
            raise ConfigurationException("Initial state is not set")

        init_state = jnp.array(self._initial_state, dtype=jnp.complex128)

        if self._model is not None:
            eom_func = self._model.get_value
        else:
            raise ConfigurationException("No equation of motion is configured.")
        psis = [init_state]

        for ti in range(1, len(time)):
            times, dt = self._construct_times(time, ti)
            psis_t = psis[ti - 1]
            eom = eom_func(times + dt / 2) * dt
            psis_t = self._propagate_in_time(psis_t, eom, jnp.arange(0, len(times), 1))
            psis.append(psis_t)

        psis_arr = jnp.array(psis)
        # if open system convert back the vectorized density matrices to matrix shape
        if self.is_open:
            dim = int(jnp.sqrt(eom.shape[-1]))
            psis_arr = vmap(ScipyExpm._convert_vec_to_dm, in_axes=(0, None))(psis_arr, dim)
        return psis_arr

    # TODO: implement get_collapseops method



"""Test GOAT over GRAPE. This is same as the example OC_GOAToverGRAPE_TLS"""

from collections.abc import Callable
from functools import partial

import jax.numpy as jnp
import numpy as np
import numpy.testing as testing
import pytest
from jax import jit
from jax.scipy.special import erf

from paraqeet.measurement.goat_over_grape import GOATOverGRAPE
from paraqeet.measurement.state_transfer_fidelity import (
    StateTransferFidelityGRAPE,
)
from paraqeet.model.closed_system import ClosedSystem
from paraqeet.model.rotating_frame_drive import RotatingFrameDrive
from paraqeet.optimization_map import OptimizationMap
from paraqeet.optimizers.scipy_optimizer_gradient import ScipyOptimizerGradient
from paraqeet.propagation.scipy_expm_grape import ScipyExpmGRAPE
from paraqeet.quantity import Array, Quantity
from paraqeet.signal.envelopes import Envelope
from paraqeet.signal.pwc_generator import PWCGenerator
from tests.model.spin_rwa import SpinRWA

T_FINAL = 20e-9
TLIST = jnp.linspace(0, T_FINAL, 26)


class FlatTopGaussianEnvelope(Envelope):
    """A flat-top Gaussian envelope."""

    def __init__(
        self,
        amplitude: Quantity,
        t_up: Quantity,
        t_down: Quantity,
        ramp_time: Quantity,
    ):
        self._amplitude = amplitude
        self._t_up = t_up
        self._t_down = t_down
        self._ramp_time = ramp_time

        self._gradient_function: Callable | None = None
        self._grad_arg_nums: tuple[int, ...] = ()

    def get_parameters(self):
        """Get all parameters of the system."""
        return [self._amplitude, self._t_up, self._t_down, self._ramp_time]

    @partial(jit, static_argnums=(0,))
    def _evaluate(self, amp: Array, t_up: Array, t_down: Array, ramp_time: Array, t: Array):
        ramp_up = 1 + erf((t - t_up) / ramp_time)
        ramp_down = 1 + erf((-t + t_down) / ramp_time)
        return jnp.squeeze(amp * ramp_up * ramp_down / 4)

    def get_value(self, t: Array | float) -> Array:
        """Compute pulse shape."""
        amp = self._amplitude.get_value()
        t_up = self._t_up.get_value()
        t_down = self._t_down.get_value()
        ramp_time = self._ramp_time.get_value()
        # returns JitWrapped
        return self._evaluate(amp, t_up, t_down, ramp_time, t)  # type: ignore


@pytest.fixture
def tone():
    tone = FlatTopGaussianEnvelope(
        amplitude=Quantity(jnp.pi / T_FINAL / 3, -jnp.pi / T_FINAL, jnp.pi / T_FINAL, name="Amplitude"),
        t_up=Quantity(1e-9, 0.0, T_FINAL, name="t_up"),
        t_down=Quantity(T_FINAL - 1e-9, 0.0, T_FINAL, name="t_down"),
        ramp_time=Quantity(2e-9, 0.5e-9, T_FINAL, name="ramp_time"),
    )
    return tone


@pytest.fixture
def gen(tone):
    gen = PWCGenerator(envelopes=[tone], tlist=TLIST)
    gen.multiply_flat_top = True

    params = gen.get_parameters()
    params[0].set_limits(-200e6, 200e6)
    params[1].set_limits(-200e6, 200e6)
    return gen


@pytest.fixture
def model(gen):
    drive = RotatingFrameDrive(gen)
    spin = SpinRWA(drives=[drive])
    model = ClosedSystem(spin)
    return model


@pytest.fixture
def prop(model):
    prop = ScipyExpmGRAPE(model, resolution=1e9)

    init = jnp.array([[1.0], [0]])  # |0>
    target = jnp.array([[0.0], [1]])  # |1>

    prop.set_initial_state(init)
    prop.set_target_state(target)
    return prop


@pytest.fixture
def fid(prop):
    init = jnp.array([[1.0], [0]])  # |0>
    target = jnp.array([[0.0], [1]])  # |1>

    zeroone = StateTransferFidelityGRAPE(
        propagation=prop,
        initial_state=init,
        target_state=target,
    )
    return zeroone


@pytest.fixture
def opt_grad(tone, fid, gen, prop):
    optmap = OptimizationMap()
    optmap.add(tone)
    optmap.register_params_with_optimizables()

    goat = GOATOverGRAPE(fid, prop, generators=gen)
    opt_grad = ScipyOptimizerGradient(goat, optimization_map=optmap)
    return opt_grad


def test_can_measure(fid, gen, prop):
    fid = GOATOverGRAPE(fid, prop, generators=[gen])
    val, grad = fid.get_value_and_gradient(times=TLIST)
    assert 0 <= fid.measure(times=TLIST)
    assert 0 <= val <= 1

    value = fid.calculate_normalized_scalar(times=TLIST)
    assert 0 <= value
    testing.assert_array_less(np.zeros_like(grad), grad)


def test_goat_over_grape(opt_grad):
    res = opt_grad.optimize(times=TLIST)
    assert res.value < 1e-4



"""Class definition of the Euler propagation model."""

import jax.numpy as jnp

from paraqeet.exceptions import ConfigurationException
from paraqeet.model.equation_of_motion import EquationOfMotion
from paraqeet.propagation.propagation import Propagation
from paraqeet.quantity import Array, Quantity


class Euler(Propagation):
    r"""Simple implementation of first order Euler propagation.

    Solves the equation of motion d/dt psi(t) = F(psi(t), t)
    with a finite step size d as psi(t+d) = psi(t) + F(psi(t), t).
    The step size can be variable and is calculated from the time array that is
    passed to the propagate function.

    Parameters
    ----------
    model: Model
        Represents the equation of motion for a given Hamiltonian.

    """

    def __init__(self, model: EquationOfMotion, resolution: float = 1e9):
        # TODO: setting a default value for resolution
        super().__init__(model, resolution)

    def get_parameters(self) -> list[Quantity]:
        """Get a list of parameters of the system.

        Returns
        -------
        list[Quantity]
            List of optimizable parameters of the system.

        """
        return []

    def propagate(self, time: Array) -> Array:
        """Calculate the first order Euler propagation.

        Performs the actual propagation calculation.

        Parameters
        ----------
        time: Array
            Vector of time samples.

        Returns
        -------
        Array
            Results of the Euler propagation.

        """
        if len(time) < 2:
            raise ValueError("Euler.propagate needs at least two time points.")

        if self._initial_state is None:
            raise ConfigurationException("Initial state is not set")
        if self._model is not None:
            equations_of_motion = self._model.get_value(time)
        else:
            raise ConfigurationException("No equation of motion is configured.")

        dt = time[1:] - time[0:-1]
        states = [self._initial_state]
        for i in range(len(dt)):
            states.append(states[-1] + dt[i] * equations_of_motion[i] @ states[-1])

        return jnp.array(states)  # Jax arrays are immutable, so listing and then packing for return



"""Test the Runge-Kutta propagation model."""

import numpy as np
import pytest

from paraqeet.propagation.runge_kutta import RungeKutta
from tests.model.dummy_model import DummyEquationsOfMotion
from tests.model.empty_hamiltonian import EmptyHamiltonian
from tests.propagation.test_common_propagation import needs_initial_state


@pytest.fixture
def rk():
    """Return a Runge-Kutta model generating method."""

    def _method(dimension):
        return RungeKutta(DummyEquationsOfMotion(EmptyHamiltonian(dimension)))

    return _method


def test_parameters(rk):
    """Test parameters of the model."""
    assert rk(2).get_parameters() == []


def test_state_dimension(rk, ts, random_state):
    """Test the state vector dimensions.

    Dimension and norm of state vectors should be the same after propagation.
    """
    dim = np.random.randint(1, 100)
    state = random_state(dim)
    runge_kutta = rk(dim)
    runge_kutta.set_initial_state(state)
    propagated_states = runge_kutta.propagate(ts)
    assert len(propagated_states) == len(ts)
    assert propagated_states[-1].shape == state.shape


def test_initial_state(rk, ts):
    """Test whether the initial state is set.

    Raises
    ------
    ConfigurationException
        If the initial state is not set.

    """
    for dim in range(2, 20):
        propagation = rk(dim)
        needs_initial_state(propagation, dim)


def test_time_steps(rk, random_state):
    """Test the Runge-Kutta time steps for propagation.

    Raises
    ------
    ValueError
        If the Runge-Kutta propagation does not get at least two time steps.

    """
    time = np.array([0])
    dim = np.random.randint(1, 100)
    state = random_state(dim)
    runge_kutta = rk(dim)
    runge_kutta.set_initial_state(state)
    with pytest.raises(
        ValueError,
        match="RungeKutta.propagate needs at least two time steps",
    ):
        runge_kutta.propagate(time)



"""Class definition for the Optimizable model."""

from abc import ABC, abstractmethod

from paraqeet.quantity import Quantity


class Optimizable(ABC):
    """Optimizable parameter model.

    This interface must be implemented by any class that provides optimizable
    parameters. The optimizer will collect all parameters (by reference) and
    update their values.

    Parameters
    ----------
    _name : str | None
        The name of the object.
    _optimizable_parameters : list[Quantity] = []
        The optimizable parameters associated with the object.
    _all_optimizable_parameters: list[Quantity] = []
        All the optimizable parameters considered in the optimization. This
        is needed to set correctly the size of the gradient and its ordering.

    """

    _name: str | None = None
    _optimizable_parameters: list[Quantity] = []
    _all_optimizable_parameters: list[Quantity] = []

    @abstractmethod
    def get_parameters(self) -> list[Quantity]:
        """Return all parameters of this class that can be optimized.

        Raises
        ------
        NotImplementedError
            Subclasses derived from this class must implement this method.

        """
        pass

    @property
    def name(self) -> str | None:
        """Get the name of the parameter.

        Returns
        -------
        str | None
            Name of the parameter.

        """
        return self._name

    @name.setter
    def name(self, name: str | None) -> None:
        """Set the name of the parameter.

        Parameters
        ----------
        name : str | None
            Value of the name to be set.

        """
        self._name = name

    @property
    def optimizable_parameters(self) -> list[Quantity]:
        """Get the optimizable parameters

        Returns
        -------
        list[Quantity]
            The list of optimizable parameters associated with the object.
        """
        return self._optimizable_parameters

    @property
    def all_optimizable_parameters(self) -> list[Quantity]:
        """Get the optimizable parameters

        Returns
        -------
        list[Quantity]
            The list of all the optimizable parameters considered in the optimization
        """
        return self._all_optimizable_parameters

    def __repr__(self):
        """Magic method for human readable representation."""
        return self.__str__()

    def __str__(self):
        """Magic method for human readable string representation."""
        return self._name or str(self.__class__)

    def set_optimizable_parameters(self, params: list[Quantity]) -> None:
        """Set which parameters associated with the object shall be considered
        during optimization.

        All quantities that are not in the response of get_parameters will
        be filtered out. This function is called by the optimizer before
        gradient based optimization to tell the layers which gradients to
        compute.

        Parameters
        ----------
        params: list[Quantity]
            List of optimizable parameters to be set.

        """
        own_params = self.get_parameters()
        self._optimizable_parameters = [p for p in params if any([p is q for q in own_params])]

    def set_all_optimizable_parameters(self, all_params: list[Quantity]) -> None:
        """Set all optimizable parameters in the optimization.

        Parameters
        ----------
        params : List[Quantity]
            List of optimizable parameters to be set.

        """
        self._all_optimizable_parameters = all_params

    def _is_optimized(self, param: Quantity) -> bool:
        """Check if a parameter is being optimized.

        Should therefore be included in gradients.

        Parameters
        ----------
        param: Quantity
            Input parameter to be checked for whether it is optimized.

        Returns
        -------
        bool
            True if parameter is optimized.

        """
        return id(param) in [id(opt_param) for opt_param in self._optimizable_parameters]
